module DM::ProfileBuilder3

  ###############################
  #dictionaries
  MAIN_DICTIONARY = 'ProfileBuilder'
  #keys
  DEFAULT_JUNCTION_KEY = 'defaultJunction'
  DEFAULT_EXTRUDE_KEY = 'defaultExtrude'
  BUILD_COMPONENT_KEY = 'buildAsComponent'
  SHOW_PATH_EDIT_TIP_KEY = 'showPathEditTip'
  SHOW_HEIGHT_WARNING_KEY = 'showHeightWarning'

  HOME_PROFILE_PATH = 'HomeProfilePath'
  HOME_ASSEMBLY_PATH = 'HomeAssemblyPath'
  LAST_PROFILE_PATH = 'LastProfilePath'
  LAST_ASSEMBLY_PATH = 'LastAssemblyPath'
  #constants
  FALL_BACK_PROFILE_PATH = File.join(File.dirname(__FILE__), 'Libraries', 'Profile Examples')
  FALL_BACK_ASSEMBLY_PATH = File.join(File.dirname(__FILE__), 'Libraries', 'Assembly Examples')
  ###############################


  this_dir = File.dirname(__FILE__)
  PATH = this_dir

  entries = Dir.entries(PATH)

  # Check if there are .rbe files in this folder - if that is the case, this
  # must be a public release so remove any old .rb and .rbs files
  if entries.find{ |file| file[-4..-1] == ".rbe" }
    Dir.foreach(PATH) do |file|
      File.delete(File.join(PATH, file)) if (file[-4..-1] == ".rbs")
    end
  end

  SOURCE_FILES = %w(
    mixins
    singleton
    PBAPI/loader
    observers
    profilebuilder
    preferences
    pathtools
    profile_library
    converter
    assembly_dialog
    lathe
    trim
    cut_openings_setting_window
    cut_openings_tool
    remove_hole_tool
    extend
    assembly_tool
    picker
    progressbar
    browser
    finder
    mark_numbers
  )

  def self.reload
    original_verbose = $VERBOSE
    $VERBOSE = nil
    api_files = %w(
      pbapi
      geometry
      path_extruder
      curve
      arc
      chain
      profile
      hole
      profilemember
      assembly
      material
      revolver
      previewer
      classify
    )
    api_files.each { |source| load(File.join(PATH, 'PBAPI', source + ".rb")) }
    #["extend"].each { |source| load(File.join(PATH, source + ".rb")) }
    (["assembly_dialog", "assembly_tool", "picker", "window", "mixins", "cut_openings_setting_window", "mark_numbers", "progressbar", "preferences", "browser"])
        .each { |source| load(File.join(PATH, source + ".rb")) }
  ensure
    $VERBOSE = original_verbose
  end

  def self.load_source_files
    SOURCE_FILES.each { |source| Sketchup::require(File.join(PATH, source))}
  end

  self.load_source_files
  PLUGIN::PBAPI::Profile.default_junction_style = PLUGIN::Preferences.junction_style_default #set default extrude mode
  PLUGIN::PBAPI::Profile.default_extrude_mode = PLUGIN::Preferences.extrude_mode_default #set default junction style

  # TODO: Freeze Modules prior to public release!!!!
  freeze
  #UI.start_timer(2.0, false) { AssemblyDialog.create.show }

end


