# License Module Injection Start
module DM::ProfileBuilder3
LIC_FILE = 'pb3.lic'.freeze
LIC_DIR = 'ProfileBuilder3'.freeze


# This method is generated by 'build_license.rb'
def self.load_c_extensions
  dir = File.dirname(__FILE__)
  su_version = Sketchup.version.to_i
  ruby_ver = case su_version
               when 21
                 '2.7'
               when 19, 20
                 '2.5'
               when 18, 17
                  '2.2'
               when 16
                 '2.0'
               else
                 '2.7'
               end

  if Object::RUBY_PLATFORM =~ /darwin/i
    live_so_path = File.join(dir, 'so', 'bundle', ruby_ver, 'MSS_Licensing.bundle')
    source_so_path = File.join(dir, 'so', 'bundle', ruby_ver, 'MSS_Licensing_source.bundle')
    if !File.exists?(live_so_path) || !FileUtils.compare_file(source_so_path, live_so_path)
      File.unlink(live_so_path) if File.exists?(live_so_path)
      FileUtils.cp(source_so_path, live_so_path)
    end
  else
    arch = Sketchup.respond_to?(:is_64bit?) && Sketchup.is_64bit? ? '64bits' : '32bits'
    so_path = File.join(dir, 'so', ruby_ver, arch)
    source_so_path = File.join(so_path, 'MSS_Licensing_source.so')
    live_so_path = File.join(so_path, 'MSS_Licensing.so')
    #copy source c++ code to live (this is needed so users can update the plugin)
    #If we only have the source file it cannot be overwritten while SketchUp is in use
    if !File.exists?(live_so_path) || !FileUtils.compare_file(source_so_path, live_so_path)
      File.unlink(live_so_path) if File.exists?(live_so_path)
      FileUtils.cp(source_so_path, live_so_path)
    end
  end
  valid_checksums = {
    'DM_ProfileBuilder3/so/2.2/32bits/MSS_Licensing_source.so' => '38b9d1690ed24a9b5ea6dbbd6feaa42f293408c89dff61b01a0830af8be2fc0f',
'DM_ProfileBuilder3/so/2.2/64bits/MSS_Licensing_source.so' => 'db37d875ace632fa867141e57e287d6fe5d4ee4d8b71a6677bfc0443b9392535',
'DM_ProfileBuilder3/so/2.5/64bits/MSS_Licensing_source.so' => 'd2ce2b90f70b03d15b8fab8caa715bedc49c32ea4a464ed6a1c776a2eb7f1f37',
'DM_ProfileBuilder3/so/2.7/64bits/MSS_Licensing_source.so' => 'b8248f63c5dc507613f20562ae9541353ea793cbde77e580bf9cfaa20bdc66b8',
'DM_ProfileBuilder3/so/bundle/2.2/MSS_Licensing_source.bundle' => '25171e65798132bae9cc8e656d9e135f001f5d8c034f29223d0ece91466206f7',
'DM_ProfileBuilder3/so/bundle/2.5/MSS_Licensing_source.bundle' => '1b212487ff0c4e9da8bf51a8c15a286c1bb1725b6f3ae329fbb1294cf214c6be',
'DM_ProfileBuilder3/so/bundle/2.7/MSS_Licensing_source.bundle' => '5ea788cd5598775980c3c9ab229cbdff17d79f87c6838a18254ce8b63e6b1d84'
  }
  cs = nil
  valid_checksums.each { |key, val|
    if source_so_path.downcase.include?(key.downcase)
      cs = Digest::SHA256.file(source_so_path).hexdigest
      raise 'ChecksumError' unless cs == cs
    end
  }
  raise 'ChecksumError' unless cs # Make sure matching checksum was calculated
  require live_so_path #load c++ code
  @did_load_extensions = true # Flag that the plugin can check for to determine whether to create the UI
  rescue Errno::EACCES => e
    UI.messagebox "Failed to load #{PLUGIN_NAME}. Please ensure the most recent version is installed, close all SketchUp windows and restart to finish the update."
  rescue Exception => e
    if e.message == 'ChecksumError'
      # Should only occur if plugin was hacked so throw an obscure message with no console output
      UI.messagebox "Failed to load #{PLUGIN_NAME}. Please close all SketchUp windows and restart."
      @did_load_extensions = false
    else
     raise
    end
end

require 'time'
require 'digest'
require 'fileutils'
require 'logger'
require 'stringio'
require 'tmpdir'
require 'net/http'
require 'uri'
require 'open-uri'
require 'json'

# This exception is raised and caught by the PLUGIN when the c++ license check fails
class InvalidLicenseException < RuntimeError; end

# RUBY 2.5 removed this function: https://github.com/rails/rails/pull/31462/files
# https://medium.com/@maciejmensfeld/ruby-2-5-0-upgrade-remarks-100d7c0dd73d
# Generates log files like: name-YYYYMMDD-9004-bxfuvf.log
def self.tmpname(name)
  t = Time.now.strftime("%Y%m%d")
  "#{name}-#{t}-#{$$}-#{rand(0x100000000).to_s(36)}"
end

# This will define a standard 'check_license' method that is exposed to Ruby
# Note that this does an offline check online and should only be used in addition
# to the other checks implemented inside c++
# Call this from a plugin by using 'PLUGIN.check_license'.  Use this check to prevent
# a dialog from opening or a tool from starting
def self.check_license
  lic = Licensing::License.new
  state = lic.get_state
  case state
  when Licensing::LICENSED, Licensing::TRIAL
    return true
  when Licensing::TRIAL_EXPIRED
    Licensing.on_trial_expired
    return false
  when Licensing::LICENSE_EXPIRED
    Licensing.on_license_expired
    return false
  else
    Licensing.on_invalid_license_found
    return false
  end
end

module Http

  # Defines a simple method that can be used to check if user is online along
  # with success and error code blocks
  def self.ping(on_success, on_error)
    if Sketchup.version.to_i < 19
      # Ping our server to test if user is online.  Note that on SU2018
      # the Sketchup.is_online method no longer works and for Mac, SU2017 and SU2018,
      # calling Sketchup.is_online inside an htmldialog callback causes SketchUp
      # to permanently hang, hence the logic below
      @ping_req = Request.new("https://0.0.0.0/")
      @ping_req.start do |req, resp|
        if resp.status_code != 0 # online
          on_success.call
        else
          # no connection
          on_error.call
        end
      end
    else
      if Sketchup.is_online
        on_success.call
      else
        on_error.call
      end
    end
  end

  class Request

    attr_accessor :body, :method


    def initialize(url, method = 'GET')
      @url = url
      @method = method
      uri = URI.parse(url)
      @host = uri.host
      @path = uri.path
      @query = uri.query
      @port = uri.port
      @scheme = uri.scheme
      @body = ""
      @req = create_request
    end

    def create_request
      Sketchup::Http::Request.new(@url, @method)
    end

    def set_download_progress_callback(&block)
      @req.set_download_progress_callback(&block)
    end

    def body=(value)
      @body = value
    end

    def cancel
      @req.cancel
    end

    def headers
      @req.headers
    end

    def headers=(hash)
      @req.headers = hash
    end

    def start(&callback)
      @req.start(&callback)
    end

    def url
      @url
    end

  end

  class Response
    attr_accessor :body, :headers, :status_code, :status_message
  end

end

module Licensing

  # NOTE: The language handler must initialized in the plugin elsewhere
  # This is just to prevent crashes where localization may not yet implemented in a particular
  # plugin that is using this license module
  LH = DMLanguageHandler.new("license.strings") unless defined?(LH)

  # NOTE: The constant 'PLUGIN' must be defined elsewhere and must
  # reference the parent module (which should always be the containing
  # module for this plugin. eg: DM::Artisan)

  # NOTE: The constant 'PLUGIN_NAME' must also be defined elsewhere and should reference
  # the standard display name for the plugin. (eg. 'Artisan' or 'Profile Builder 2')

  # NOTE: The constant 'PLUGIN::PATH' must be defined elsewhere and should reference
  # the folder that contains the main ruby files for this plugin

  # NOTE: The constant 'PLUGIN::VERSION' must be defined elsewhere and should reference
  # the current version number for this plugin (eg. 1.2.3)

  # NOTE: The constant 'PLUGIN::LIC_DIR' must be defined elsewhere and should reference
  # the directory name for the license file for this plugin

  # NOTE: The constant 'PLUGIN::LIC_FILE' must be defined elsewhere and should reference
  # the name for the license file for this plugin (eg. pb2.lic)

  ON_MAC = Sketchup.platform == :platform_osx
  SHARED_LOCATION = ON_MAC ? File.join(ENV['HOME'], 'Library', 'Application Support') : ENV['ALLUSERSPROFILE']
  MAC_LOCATION = ENV['systemroot'] ? File.join(ENV['systemroot'], 'System32') : '' #systemroot is a read-only variable so we will use it instead of windir
  require 'win32ole' unless ON_MAC

  def self.grant_full_access_windows(file_path) #grants everyone full access to file or folder
    if !ON_MAC #windows only
      params = File.exists?(MAC_LOCATION) ? "/c cd #{MAC_LOCATION} && " : '/c ' #change to normal icacls.exe location if folder exists
      params += "icacls \"#{file_path}\" /GRANT *S-1-1-0:(OI)(CI)(F)" #grant EVERYONE: folder, sub-folders, files, full access
      shell = WIN32OLE.new('Shell.Application')
      shell.ShellExecute('cmd.exe', params, '', 'open', 0)
    end
  end

  APP_DATA = File.join(SHARED_LOCATION, LIC_DIR)
  if !FileTest.exist?(APP_DATA) #file/folder does not exist
    Dir.mkdir(APP_DATA)  #create folder
    Licensing.grant_full_access_windows(APP_DATA) #grant all users full access
  end
  LOG_FILES = File.join(APP_DATA, 'Logs')

  DEBUG = 0   # Low-level information, mostly for developers.
  INFO = 1    # Generic (useful) information about system operation.
  WARN = 2    # A warning.
  ERROR = 3   # A handleable error condition.
  FATAL = 4   # An unhandleable error that results in a program crash.
  UNKNOWN = 5 # An unknown message that should always be logged.

  module Log
    # Logs will be saved into a sub folder in the license directory
    if File.exist?(APP_DATA)
      @logs_path = LOG_FILES
    else
      UI.messagebox(LH['Unable to find AppData directory.'])
      @logs_path = nil # Indicates error.
    end


    def self.create_log_file(logs_path)
      if logs_path.nil?
        raise(RuntimeError, 'No path to log files.') # This will be caught by the rescue at the bottom.
      end

      # Make sure the destination directory exist.
      unless File.directory?(logs_path)
        FileUtils.mkdir_p(logs_path) #create any of the folders that are missing
        Licensing.grant_full_access_windows(logs_path) #grant all users full access
      end

      if @log_file
        log_filename = @log_file.path #re-open the previously used log file
      else
        log_basename = PLUGIN.tmpname('log-') # Generates log files like: log-20141106-2652-1dpjrlb.log
        log_filename = File.join(logs_path, "#{log_basename}.log")
      end

      # Create the log file.
      mode = File::WRONLY | File::APPEND | File::CREAT #Open the file in write mode, append to existing file, create if needed
      log_file = File.open(log_filename, mode)

      log_file
    rescue
      # Account for permission failure and other errors. Fall back to logging
      # to a StringIO object. It won't be logged to file, but the extension can
      # still work.
      return StringIO.new
    end

    # Thread safe logger.
    # http://www.ruby-doc.org/stdlib-2.0.0/libdoc/logger/rdoc/Logger.html
    @logger = nil

    # Create an accessors that perform lazy initialization of the log file so
    # we only create a log file when the extension is ever being used.
    def self.logger
      @log_file = self.create_log_file(@logs_path) #creates or opens the log file
      @logger ||= Logger.new(@log_file) #set logger only if it is not already set
      @logger
    end

    def self.end_writing
      self.logger.close
      @logger = nil
    end

    # http://www.ruby-doc.org/stdlib-2.0.0/libdoc/logger/rdoc/Logger/Severity.html
    # DEBUG (0)   # Low-level information, mostly for developers.
    # INFO (1)    # Generic (useful) information about system operation.
    # WARN (2)    # A warning.
    # ERROR (3)   # A handleable error condition.
    # FATAL (4)   # An unhandleable error that results in a program crash.
    # UNKNOWN (5) # An unknown message that should always be logged.
    def self.write(string, type = DEBUG)
      array = [string.to_s] #make it a one element array so we can pack and unpack it
      encoded = array.pack('m')
      case(type)
      when DEBUG
        self.logger.debug{ encoded }
      when INFO
        self.logger.info{ encoded }
      when WARN
        self.logger.warn{ encoded }
      when ERROR
        self.logger.error{ encoded }
      when FATAL
        self.logger.fatal{ encoded }
      when UNKNOWN
      else
      end
      self.end_writing
    end

    def self.debug(string)
      self.write(string, type = DEBUG)
    end

    def self.error(string)
      self.write(string, type = ERROR)
    end

    def self.fatal(string)
      self.write(string, type = FATAL)
    end

    def self.info(string)
      self.write(string, type = INFO)
    end

    def self.warn(string)
      self.write(string, type = WARN)
    end

  end

  def self.open_log_file_location
    path = "file://#{LOG_FILES}"
    path.gsub!(" ", "%20") if (ON_MAC && Sketchup.version.to_i >= 19)
    UI.openURL(path)
  end

  # Action to be taken if an trial license is expired
  def self.on_trial_expired
    Sketchup.active_model.abort_operation # Should be safe to call this even if there is no current operation
    UI.messagebox(TRIAL_EXPIRED_MESSAGE)
    Sketchup.active_model.select_tool(nil)
    LicenseInfo.instance.show
  rescue
    # We want to make sure this operation will return since it may have been called from the c++
    # online license check block and we need to ensure that block completes execution
    return
  end

  # Action to be taken if an invalid license is found
  # @param [InvalidLicenseException] err sub-class of RuntimeError
  def self.on_invalid_license_found(err = nil)
    Sketchup.active_model.abort_operation # Should be safe to call this even if there is no current operation
    UI.messagebox(INVALID_LICENSE_MESSAGE)
    Sketchup.active_model.select_tool(nil)
    LicenseInfo.instance.show
  rescue
    # We want to make sure this operation will return since it may have been called from the c++
    # online license check block and we need to ensure that block completes execution
    return
  end

  # Action to be taken if subscription license is expired
  def self.on_license_expired
    Sketchup.active_model.abort_operation # Should be safe to call this even if there is no current operation
    UI.messagebox(LICENSE_EXPIRED_MESSAGE)
    Sketchup.active_model.select_tool(nil)
    LicenseInfo.instance.show
  rescue
    # We want to make sure this operation will return since it may have been called from the c++
    # online license check block and we need to ensure that block completes execution
    return
  end

  PLIST_LOCATION = File.join(SHARED_LOCATION, 'MSS')
  PLIST_FILE_NAME = 'active.plist'

  # Licensing LH messages
  SERVER_ERROR_MESSAGE =  LH["Unable to validate license. Please try again later or contact customer support for plugin: "] + PLUGIN_NAME
  ACTIVATIONS_EXCEEDED_MESSAGE = LH["You have exceeded the allowable number of activations for this license. Please remove a previously activated license from another computer and then try again."]
  VALIDATION_MESSAGE = LH["There was an issue validating your license. Please contact support and send your log files: "] + LOG_FILES #used for when validating a license fails
  EXCEEDED_ACTIVATION_ERROR = LH["You have exceeded the allowable number of activations for this license. Please contact customer support. Send a detailed description and your log files: "] + LOG_FILES
  LICENSE_FOUND_MESSAGE = LH["License found. Thank you for purchasing!"]
  TRIAL_FOUND_MESSAGE = LH["Trial License found."]
  SYSTEM_ID_INVALID = LH["System ID does not match expected format."]
  SYSTEM_ID_CHANGED = LH["System ID has changed."]
  ONLINE_LICENSE_CHECK_FAILED = LH["There was a problem validating your plugin license. Please re-activate your license for: "] + PLUGIN_NAME
  OFFLINE_REMOVE_LIC_MESSAGE = LH["You must be connected to the internet to remove your license."]
  OFFLINE_RENEW_LIC_MESSAGE = LH["You must be connected to the internet to renew your license."]
  TRIAL_EXPIRED_MESSAGE = LH["Your trial has expired for plugin: "] + PLUGIN_NAME
  INVALID_LICENSE_MESSAGE = LH["A valid license could not be found for plugin: "] + PLUGIN_NAME
  LICENSE_EXPIRED_MESSAGE = LH["Your license has expired. Please use the License Manager to renew your license for plugin: "] + PLUGIN_NAME
  TRIAL_ABUSE_MESSAGE = LH["Sorry, you have exceeded the allowable number of trial activations. Please purchase a license to keep using this product."]

  module HtmlDialogHelper
    def add_common_callbacks
      @dlg.add_action_callback("ready") {
        localize
        refresh_dialog
        disable_actions
      }
    end

    def show
      if @dlg && @dlg.visible?
        @dlg.bring_to_front
      else
        init_dialog
        add_action_callbacks # Window is being opened, NEED to add callback since HtmlDialogs lose connection when closed
        @dlg.show
        @dlg.center
      end
    end

    def add_slashes(text)
      val = ""
      if text
        val = text.dup
        val.gsub!(/\\/, "\\\\\\")
        val.gsub!(/\n/, "\\\\n")
        val.gsub!(/"/, "\\\"")
      end
      val
    end

    # Sets the value of an input field in the Html dialog
    def set_value(id, value, dlg = self)
      new_value = add_slashes(value.to_s)
      cmd = "document.getElementById('#{id}').value=\"#{new_value}\""
      dlg.execute_script(cmd)
    end

    def call_js(function, *arguments)
      js_args = arguments.map { |x| x.to_json }.join(", ")
      javascript = "#{function}(#{js_args});"
      @dlg.execute_script(javascript)
    end

    def localize
      call_js("localize", LH.strings)
    end

    def disable_actions
      call_js("disableContextMenu")
      call_js("disableSelect")
    end

  end # Htmldialog helper

  class LicenseInfo

    @@info = nil
    include Licensing::HtmlDialogHelper

    def self.instance
      @@info = new unless @@info
      return @@info
    end

    # PLUGIN set to plugin module (Artisan, PB2 ...)
    def initialize
      @dlg = nil
    end # init

    # Always initialize the dialog each time it is shown (needed to avoid dialog issues on Mac)
    def init_dialog
      dlg_name = LH["License Manager: "] + PLUGIN_NAME
      dlg_id = "#{PLUGIN_NAME}LicInfo"
      height = ON_MAC ? 320 : 330
      @dlg = UI::HtmlDialog.new(
          dialog_title: dlg_name,
          preferences_key: dlg_id,
          scrollable: false,
          resizable: true,
          width: 450,
          height: height,
          left: 150,
          top: 150
      )
      @dlg.set_file(File.join(PLUGIN::PATH, 'html', 'lic_info.html'))
    end

    # HtmlDialogs need to have their callbacks added any time they are opened as the connection is removed when the dialog is closed
    def add_action_callbacks
      add_common_callbacks
      @dlg.add_action_callback('ValueChanged') { |_, p|
        pair = p.split("|")
        id = pair[0]
        case id
        when "addLicense"
          if @lic.get_state == LICENSED
            UI.messagebox(LH["You already have a valid license."])
          else
            @lic.register {refresh_dialog}
          end
        when "renewLicense"
          # Note that the Renew License button should only be visible if the user has a paid or NFR license
          # and the license is in an expired state
          @lic.get_state
          if @lic.lic_for == "NFR"
            # This message should only appear for partners so we won't bother localizing it.
            UI.messagebox("Please contact customer support to request a new NFR license.")
          elsif @lic.lic_for == "Commercial"
            result = UI.messagebox(LH["Are you sure you want to renew and reactivate your license on this computer?"], MB_YESNO)
            if result == IDYES
              on_valid = ->  { refresh_dialog }
              on_invalid = -> (reason){ on_invalid_license_refresh(reason) }
              make_sure = -> {}
              @lic.renew_license(on_valid, on_invalid, make_sure)
            end
          else
            UI.messagebox(LH["Your license type can not be renewed. Please purchase a new license to continue using this product."])
          end
        when "removeLicense"
          result = UI.messagebox(LH["Are you sure you want to remove your license?"], MB_YESNO)
          if result == IDYES
            on_valid = -> { on_licensed_removed } #callback function
            on_invalid = -> (reason){ on_invalid_license_removal(reason) } #callback function
            make_sure = -> {} #callback function
            @lic.remove_license(on_valid, on_invalid, make_sure)
          end
        end
      }
    end

    def on_licensed_removed
      License.remove_license_file
      message = LH["Your license has been removed."]
      UI.messagebox(message)
      refresh_dialog
    end

    def on_invalid_license_refresh(reason)
      message = LH["There are no renewals available for this license. Please ensure your administrator has purchased "\
      "a license renewal for this account.  Would you like to purchase now?"]
      result = UI.messagebox(message, MB_YESNO)
      if result == IDYES
        UI.messagebox(LH["A website will open where you can purchase a license renewal for this product. Once you have received your renewed license information, please renew your license using the License Manager."])
        @lic.renew_subscription
      end
      puts reason # TODO: should this be logged?
    end

    def on_invalid_license_removal(reason)
      License.remove_license_file
      puts reason # TODO: should this be logged?
      refresh_dialog
    end

    def refresh_dialog
      @lic = License.new
      state = @lic.get_state
      product = @lic.product + " (#{PLUGIN::VERSION})"
      set_button_visibility(state)
      set_button_widths(state)
      set_value("product", product, @dlg)
      set_value("customer", @lic.customer, @dlg)
      set_value("email", @lic.email, @dlg)
      set_value("expires", @lic.expires, @dlg)
      set_value("type", get_friendly_license_type_name(@lic.lic_type), @dlg)
      set_value("for", @lic.lic_for, @dlg)
      set_value("licensor", @lic.licensor, @dlg)
      set_value("status", get_friendly_license_state_name(state), @dlg)
    end

    # sets the visibility of the buttons depending on the state
    def set_button_visibility(status)
      case status
      when NOT_LICENSED
        cmd = "$('#removeLicense').hide();$('#addLicense').show();$('#renewLicense').hide();"
      when TRIAL, TRIAL_EXPIRED, LICENSED, LICENSE_ERROR
        cmd = "$('#removeLicense').show();$('#addLicense').hide();$('#renewLicense').hide();"
      when LICENSE_EXPIRED
        cmd = "$('#removeLicense').show();$('#addLicense').hide();$('#renewLicense').show();"
      end
      @dlg.execute_script(cmd)
    end

    # sets the width of the buttons depending on the state
    def set_button_widths(status)
      case status
      when TRIAL, TRIAL_EXPIRED, LICENSED
        # one button displayed - remove
        cmd = "$('#removeLicense').width('90%');"
      when LICENSE_EXPIRED
        # two buttons displayed - remove or renew
        cmd = "$('#removeLicense').width('160px');$('#renewLicense').width('160px')"
      else
        # display a full width activate button
        cmd = "$('#addLicense').width('90%');"
      end
      @dlg.execute_script(cmd)
    end

    # Returns a friendly and / or localized type name for the license based on the data that was found in the database
    def get_friendly_license_type_name(type)
      if PLUGIN.respond_to?(:license_type)
        return PLUGIN.license_type(type)
      end
      name = ""
      case type
      when "SUS" # Subscription licenses
        name = LH["Single User Subscription"]
      when "single user", "SUP" # Permanent licenses
        name = LH["Single User"]
      when "Trial"
        name = LH["Trial"]
      else
        name = type
      end
      name
    end

    # Returns a friendly and / or localized name for the status of the license
    def get_friendly_license_state_name(status)
      name = ""
      case status
      when LICENSED
        if @lic.expires != "permanent"
          name = LH["Active - Days Remaining: "] + "#{@lic.days_remaining}"
        else
          name = LH["Licensed"]
        end
      when LICENSE_EXPIRED
        name = LH["License Expired"]
      when TRIAL
        if @lic.expires != "permanent"
          name = LH["Trial License - Days Remaining: "] + "#{@lic.days_remaining}"
        else
          name = LH["Licensed"]
        end
      when TRIAL_EXPIRED
        name = LH["Trial Expired"]
      when NOT_LICENSED
        name = LH["Not Licensed"]
      when LICENSE_ERROR
        name = LH["License Error"]
      end
      name
    end

  end #LicenseInfo

  class License

    include Licensing::HtmlDialogHelper

    FILE = File.join(Licensing::APP_DATA, LIC_FILE)
    COMPANY = "MindSight Studios Inc."

    def self.windows_mac_addr_string
      if !ON_MAC
        params = File.exists?(MAC_LOCATION) ? "#{MAC_LOCATION}/" : '' #change to normal ipconfig.exe location if folder exists
        params += "ipconfig /all" #add the ipconfig command that will execute (either we are in the correct folder or hopefully the system path is properly set)
      end
      return params
    end

    def self.file
      FILE
    end

    #gets host id that is stored in the plist to verify that the system is licensed
    def self.get_host_id_from_plist
      file_name = File.join(PLIST_LOCATION, PLIST_FILE_NAME)
      if !File.exists?(file_name)
        Log.error("Plist file: #{file_name} does not exist cannot get valid plist host_id")
      end

      if ON_MAC
        cmd = "defaults read \"#{file_name}\" \"#{PLUGIN_NAME}\"" #reads the value from the file
        host_id = %x[#{cmd}].rstrip #result will be the host id, nil if it didn't exist or failed
      else
        hash = self.read_plist_file
        host_id = hash.nil? ? nil : hash["#{PLUGIN_NAME}"]
      end

      return host_id ? host_id : '' #do not return nil
    end

    #opens the plist file to read if it exists, then gets the host id hash
    def self.read_plist_file
      file_name = File.join(PLIST_LOCATION, PLIST_FILE_NAME)
      hash = nil
      begin
        if File.exists?(file_name)
          out = File.new(file_name, 'rb')
          text = out.read
          out.close unless out.closed?
          hash = JSON.parse(text)
        end
      rescue
        out.close if out && !out.closed?
      end

      return hash
    end

    # opens and reads a file into an array for each line
    def self.read_file_to_array(file_path)
      f = File.new(file_path, "rb")
      lines = f.read.split("\n")
      lines.each { |line| line.strip! } # ensures \r character will also get removed (remnant from older license files)
      f.close
      lines
    end

    # return true if file can be written to
    def self.writable(file_path)
      return true unless Object.const_defined?("Pathname") # For SU 2016 Pathname is not valid so returning true
      foo = Pathname.new(file_path)
      foo.writable?
    end

    # creates the license file
    def self.write_file(file_path, data, serial, sig)
      out = File.new(file_path, "wb")
      out.puts "LICENSE\n"
      out.puts data + "\n"
      out.puts serial + "\n"
      out.puts sig + "\n"
      out.close unless out.closed?
    end

    #removes or overwrites the host id for the Plugin from the file
    def self.remove_host_id_from_plist
      file_name = File.join(PLIST_LOCATION, PLIST_FILE_NAME)
      if ON_MAC
        cmd = "defaults delete \"#{file_name}\" \"#{PLUGIN_NAME}\"" #deletes the key, value from the plist
        result = system(cmd) #result will be true if cmd was success and false otherwise
        #can check if it fails but that should mean the file or entry was already removed
      else
        begin
          if FileTest.exist?(file_name)
            hash = License.read_plist_file
            if hash.nil?
              File.delete(file_name)
            else
              hash.delete("#{PLUGIN_NAME}") #remove the key and therefore its value from the hash
              out = File.new(file_name, 'wb')
              out.puts(hash.to_json) #save new reduced plist json hash
              out.close unless out.closed?
            end
          end
        rescue Exception => e
          Log.error("Problem when removing: #{PLUGIN_NAME} entry from plist")
          Log.error(e.message)
          Log.info(e.backtrace)
            #todo UI message the user?
        ensure
          out.close if out && !out.closed?
        end
      end
    end

    def self.remove_license_file
      if FileTest.exist?(FILE)
        File.delete(FILE)
      end
    end

    def initialize
      @licensor = "" #needed so it can be displayed by LicenseInfo class
      @product = "#{PLUGIN_NAME}" #needed so it can be displayed by LicenseInfo class
      @customer = LH["Unregistered User"] #needed so it can be displayed by LicenseInfo class
      @email = "" #needed so it can be displayed by LicenseInfo class
      @type = "" #single user, network, trial #needed so it can be displayed by LicenseInfo class
      @lic_for = "" #trial, commercial, educational, nfr #needed so it can be displayed by LicenseInfo class
      @expires = "" #needed so it can be displayed by LicenseInfo class
      @days_remaining = 0 #needed so it can be displayed by LicenseInfo class
      @host_id = "" #the selected host_id #needed so it can be written to plist
      @callback = nil
      @dlg = nil
    end

    # Always initialize the dialog each time it is shown (needed to avoid dialog issues on Mac)
    def init_dialog
      dlg_name = LH["Activate License: "] + PLUGIN_NAME
      dlg_id = "Reg#{PLUGIN_NAME}"
      height = ON_MAC ? 165 : 175
      @dlg = UI::HtmlDialog.new(
          dialog_title: dlg_name,
          preferences_key: dlg_id,
          scrollable: false,
          resizable: true,
          width: 500,
          height: height,
          left: 150,
          top: 150
      )
      rb_initialize # Sets the activation url
    end

    # HtmlDialogs need to have their callbacks added any time they are opened as the connection is removed when the dialog is closed
    def add_action_callbacks
      return if @dlg.visible?

      @dlg.add_action_callback("create_license") { |_, data|
        cb_create_license(data)
      }
      @dlg.add_action_callback("lic_not_found") {
        @dlg.close
        if ON_MAC
          text = LH["Sorry, your license key does not appear to be valid.  Please try again and be sure to use copy / paste (CMD+C and CMD+V) to enter the key from your license information email."]
        else
          text = LH["Sorry, your license key does not appear to be valid.  Please try again and be sure to use copy / paste (CTRL+C and CTRL+V) to enter the key from your license information email."]
        end
        UI.messagebox(text)
      }
      @dlg.add_action_callback("entitlement_exceeded") {
        @dlg.close
        UI.messagebox(ACTIVATIONS_EXCEEDED_MESSAGE)
        PLUGIN.on_activations_exceeded if PLUGIN.respond_to?(:on_activations_exceeded)
      }
      @dlg.add_action_callback("set_the_product") {
        set_product
        set_host_ids
      }
      @dlg.add_action_callback("open_url") do |_, p|
        p.gsub!(" ", "%20") if (ON_MAC && Sketchup.version.to_i >= 19)
        UI.openURL(p)
      end
    end

    # Over-ride inherited Object methods to help protect class
    def extend(*args)
    end
    def inspect
      ""
    end
    def instance_eval(*args)
    end
    def instance_variables
      []
    end
    def method(*args)
    end
    def methods
      []
    end
    def private_methods
      []
    end
    def public_methods
      []
    end
    def protected_methods
      []
    end
    def respond_to?(*args)
    end
    def send(*args)
    end
    def singleton_methods
      []
    end
    def instance_methods
      []
    end
    def public_instance_methods
      []
    end
    def protected_instance_methods
      []
    end
    def private_instance_methods
      []
    end
    def to_a
      []
    end
    def to_s
      ""
    end
    def constants
      []
    end
    def class_variables
      []
    end

    public
    def register(&block)
      is_online = -> {
        @callback = block
        rb_register
      }
      is_offline = -> {
        Log.warn("Sketchup.is_online returned FALSE")
        UI.messagebox(LH["No internet connection can be found.  You must be "\
                          "connected to the internet to activate your license."])
      }
      Http.ping(is_online, is_offline)
    end

    def host_id
      @host_id.dup
    end

    def product # needed so it can be displayed by LicenseInfo class
      if PLUGIN.respond_to?(:product_name)
        # Allows the plugin to display a different product name depending on the
        # license info
        PLUGIN.product_name(self)
      else
        @product.dup
      end
    end

    def customer #needed so it can be displayed by LicenseInfo class
      @customer.dup
    end

    def email #needed so it can be displayed by LicenseInfo class
      @email.dup
    end

    def expires #needed so it can be displayed by LicenseInfo class
      @expires.dup
    end

    def serial #needed for certain functions
      @serial.dup
    end

    def lic_type #needed so it can be displayed by LicenseInfo class
      @type.dup
    end

    def lic_for #needed so it can be displayed by LicenseInfo class
      if PLUGIN.respond_to?(:lic_for)
        # Allows the plugin to over-write lic for status
        PLUGIN.lic_for(@lic_for)
      else
        @lic_for.dup
      end
    end

    def licensor #needed so it can be displayed by LicenseInfo class
      @licensor.dup
    end

    def days_remaining #needed so it can be displayed by LicenseInfo class
      @days_remaining
    end

    private
    def set_product
      set_value('productSU', PLUGIN_NAME, @dlg)
    end

    #creates the file if needed then adds the host id to it with the Plugin name as the key
    def add_host_id_to_plist
      file_name = File.join(PLIST_LOCATION, PLIST_FILE_NAME)
      if ON_MAC
        cmd = "defaults write \"#{file_name}\" \"#{PLUGIN_NAME}\" -string \"#{@host_id}\"" #creates folders and file if needed then sets key value
        result = system(cmd) #result will be true if cmd was success and false otherwise
        if result == false
          raise("Failed to add host id: #{@host_id} to the plist")
        end
      else
        Dir.mkdir(PLIST_LOCATION) unless FileTest.exist?(PLIST_LOCATION)
        Licensing.grant_full_access_windows(PLIST_LOCATION)#change permission full access to everyone
        begin
          hash = License.read_plist_file
          out = File.new(file_name, 'wb')
          hash.nil? ? hash = { "#{PLUGIN_NAME}" => @host_id } : hash["#{PLUGIN_NAME}"] = @host_id
          out.puts(hash.to_json) #save new complete plist json hash
          out.close unless out.closed?
        ensure
          out.close if out && !out.closed?
        end
      end
    end

    # This class is frozen in c++
  end

  # This method must be defined in the plugin loader.rb file prior to this file being loaded
  PLUGIN.load_c_extensions

  freeze # module Licensing

end

class AutoUpdate
  FILENAME = File.join(File.dirname(Licensing::License::FILE), "#{PLUGIN_NAME}.AutoUpdate.rbz").freeze #using lic_dir since we know this should be usable to all users
  SECTION = "DM #{PLUGIN_NAME}".freeze
  VARIABLE = 'AutoUpdate'.freeze
  AUTO_CHECK_ID = 'AutoCheck'.freeze
  @@settings = nil
  @@update_available = false
  @@checked = false
  REMOTE_PATH = "https://0.0.0.0/".freeze
  REMOTE_DIST = "https://0.0.0.0/".freeze
  include Licensing::HtmlDialogHelper

  GREATER = 100
  EQUAL = 150
  LESS = 300

  def self.instance(&block)
    if @@settings
      @@settings.set_callback{ block.call if block }
    else
      @@settings = new { block.call if block }
    end
    return @@settings
  end

  def self.auto_check_enabled
    Sketchup.read_default(SECTION, VARIABLE, 'true').to_s == 'true' ? true : false #defaults to enabled (true) also ensures we get a true or false only
  end

  def self.check_for_update(auto_check = false, &block)
    block.call if block
    return nil
  end

  #compare versions and see if this is an update
  def self.is_version_newer?(new_version)
    version_regex = /^(?<major>\d+)\.(?<minor>\d+)\.(?<other>\d+)$/
    current_v = VERSION
    if !version_regex.match(current_v)
      #puts "Your current version number is invalid"
      return false
    end
    if !version_regex.match(new_version)
      #puts "The remote version number is invalid"
      return false
    end

    major = self.compare_part(new_version, current_v, version_regex, 'major')
    case major
    when GREATER
      return true;
    when LESS
      return false;
    end

    minor = self.compare_part(new_version, current_v, version_regex, 'minor')
    case minor
    when GREATER
      return true;
    when LESS
      return false;
    end

    other = self.compare_part(new_version, current_v, version_regex, 'other')
    return true if other == GREATER

    return false
  end

  # use regex to compare different parts of 2 strings
  def self.compare_part(new, cur, regex, part)
    update = new.match(regex)[part].to_i
    current = cur.match(regex)[part].to_i
    if update > current
      return GREATER
    elsif update == current
      return EQUAL
    else
      return LESS
    end
  end

  #PLUGIN set to plugin module (Artisan, PB2 ...)
  def initialize(&block)
    @callback = block
    @dlg = nil
    @auto_check = AutoUpdate.auto_check_enabled
  end #init

  def init_dialog
    dlg_name = LH["Update: "] + PLUGIN_NAME
    dlg_id = "#{PLUGIN_NAME}AutoUpdate"
    @dlg = UI::HtmlDialog.new(
        dialog_title: dlg_name,
        preferences_key: dlg_id,
        scrollable: false,
        resizable: true,
        width: 450,
        height: 200,
        left: 150,
        top: 150
    )
    url = "#{REMOTE_PATH}update-dialog.php?plugin=#{PLUGIN_NAME}&update=#{@@update_available}"
    url.gsub!(/\s/, "%20")
    @dlg.set_url(url)
  end

  # HtmlDialogs need to have their callbacks added any time they are opened as the connection is removed when the dialog is closed
  def add_action_callbacks
    @dlg.add_action_callback('initialize') { |_, p|
      js_cmd = "document.getElementById('#{AUTO_CHECK_ID}').checked = #{@auto_check}; updateValue();"
      @dlg.execute_script(js_cmd)
      localize
      disable_actions
    }
    @dlg.add_action_callback('installUpdate') { |_, p|
      download_rbz
    }
    @dlg.add_action_callback('openChangelog') { |_, p|
      url = "#{REMOTE_DIST}update-check.php?name=#{PLUGIN_NAME}&changelog=1"
      url.gsub!(" ", "%20") if (ON_MAC && Sketchup.version.to_i >= 19)
      UI.openURL(url)
    }
    @dlg.add_action_callback('auto_check') do |_, value|
      @auto_check = value
    end
    @dlg.set_on_closed {
      Sketchup.write_default(SECTION, VARIABLE, @auto_check)
      @callback.call if @callback
      @callback = nil
    }
  end

  def set_callback(&block)
    @callback = block
  end

  def download_rbz
    filename = FILENAME
    @callback = nil # clear callback since user should restart
    url = "#{REMOTE_DIST}download.php?name=#{PLUGIN_NAME}"
    url.gsub!(/\s/, "%20")
    @rbz_req = Http::Request.new(url)
    @rbz_req.set_download_progress_callback do |current, total|
      percentage = (current.to_f / total * 100).round
      cmd = "setDownloadProgress(#{percentage})"
      @dlg.execute_script(cmd)
    end
    @dlg.set_on_closed do
      puts "RBZ Download was cancelled"
      @rbz_req.cancel
    end

    @rbz_req.start do |req, resp|
      case resp.status_code
      when 200
        file_data = resp.body
        File.open(filename, "wb:BINARY") { |temp_file|
          temp_file.write(file_data)
        }
        result = extract_rbz
        UI.messagebox(LH["The update is not finished yet. Please close all SketchUp windows and restart to load the update."]) if result
        @dlg.close
      else
        UI.messagebox(LH["Error downloading the update. Please ensure you are online."])
      end
    end
  end

  def extract_rbz
    path = FILENAME
    begin
      Sketchup.install_from_archive(path)
      return true
    rescue Interrupt => error
      UI.messagebox(LH["Update was canceled."])
    rescue Exception => error
      UI.messagebox(LH["An error occured while installing the update. Please manually download the update from the website."])
    end
    return false
  end

end

end # plugin module
# This file is generated from the build_license.rb script. 
# Do not edit this file manually.

require 'sketchup.rb'

module DM::ProfileBuilder3

  ON_MAC = ((Object::RUBY_PLATFORM =~ /mswin/i)||(Object::RUBY_PLATFORM =~ /mingw/i)) ? false : ((Object::RUBY_PLATFORM =~ /darwin/i) ? true : :other)
  DEFAULT_PROFILE = "PB2Crown 3.929203823352367 4.0 2.7964601911676112 4.0 0.0 1.2035398088323817 0.0 0.0 0.3003639209265927 0.0 0.3130053980477854 0.06209879089682957 0.3393234228461921 0.11974797428535666 0.37796248909738495 0.16997833695464948 0.42693249704232983 0.21020277102806517 0.4837112531999921 0.2383495225962946 0.6222082675369407 0.2710307808941259 0.6832675004887889 0.29659079409831435 0.7377579110604131 0.33417189785273393 0.7833457432769606 0.382164540595781 0.8180785274340963 0.43851325760299886 0.8404687019278683 0.5008047039644659 0.849557323653336 0.5663710151219199 1.0619474555371369 0.5663710151219199 1.0619474555371369 1.1170382034873434 1.456394752838186 1.2333535534081257 1.8048155173207903 1.451804696350095 2.0813536082902715 1.7561804759684136 2.2654872643695114 2.1238933091333436 2.537491855143415 2.588422103624559 2.9164607768147945 2.9707257195121635 3.3785888012859004 3.246789457282894 3.894846996621112 3.399272158146811|extended|1|false|0.0|0.0|0.0|30.0|1.0|1.0|2|Default|Layer0\n"
  require 'win32ole' unless ON_MAC

  HELP_URL = "http://profilebuilder4sketchup.com/help3"
  MAIN_REG_SECTION = "DM_ProfileBuilder"
  PROFILE_RECENT_SEC_STATE = "ProfileRecentSectionDisplay"
  ASSEMBLIES_RECENT_SEC_STATE = "AssemblyRecentSectionDisplay"
  REG_RECENT_ASSEMBLIES = "RecentAssemblyPaths"
  REG_RECENT_PROFILES = "RecentProfilePaths"
  RECENT_ARRAY_MAX_LENGTH = 5
  EDIT_PROFILE_TRANSITION = 0.75
  PICK_FROM_MODEL = LH["Pick From Model.."]

  class InvalidLicenseException < RuntimeError; end
  INVALID_LICENSE_MESSAGE = LH["A valid license could not be found."]
  TRIAL_EXPIRED_MESSAGE = LH["Your trial license for Profile Builder 3 has expired."]

  CMD_HASH = {
      "profile" => {
          name: LH["Profile Dialog"],
          tooltip: LH["Open the Profile Dialog"],
          status_bar_text: LH["Create Profiles and Build Profile Members"]
      },
      "assembly" => {
          name: LH["Assembly Dialog"],
          tooltip: LH["Open the Assembly Dialog"],
          status_bar_text: LH["Create, Edit and Build Assemblies"]
      },
      "path_select" => {
          name: LH["Smart-Path Selection"],
          tooltip: LH["Smart-Path Selection Tool"],
          status_bar_text: LH["Select complex paths with direction"]
      },
      "extend" => {
          name: LH["Extend/Split"],
          tooltip: LH["Extend or Split Profile Members"],
          status_bar_text: LH["Modify the length of Profile Members or split a member into multiple members"]
      },
      "hole_tool" => {
          name: LH["Hole Tool"],
          tooltip: LH["Hole Tool"],
          status_bar_text: LH["Create holes through any object"]
      },
      "trimsolid" => {
          name: LH["Trim to Solid"],
          tooltip: LH["Trim to Solid Tool"],
          status_bar_text: LH["Trim solid Profile Members against another Solid object"]
      },
      "trimface" => {
          name: LH["Trim to Face"],
          tooltip: LH["Trim to Face Tool"],
          status_bar_text: LH["Trim Profile Members to another Face in the model"]
      },
      "edit_path" => {
          name: LH["Edit Path"],
          tooltip: LH["Edit the Path of the active Profile Member or Assembly"],
          status_bar_text: LH["First, double-click a Profile Member or Assembly"]
      },
      "edit_profile" => {
          name: LH["Edit Profile"],
          tooltip: LH["Edit the Profile of the active Profile Member"],
          status_bar_text: LH["First, double-click a Profile Member"]
      },
      "assign_mark_numbers" => {
          name: LH["Assign Mark Numbers"],
          tooltip: LH["Assign Mark Numbers"],
          status_bar_text: LH["Batch rename Profile Members and assign numbers"]
      },
      "import_plb" => {
          name: LH["Import PLB"],
          tooltip: LH["Import PLB"],
          status_bar_text: LH["Import a PLB file from the original Profile Builder"]
      },
      "preferences" => {
          name: LH["Preferences"],
          tooltip: LH["Preferences"],
          status_bar_text: LH["General Settings"]
      },
      "help" => {
          name: LH["Help"],
          tooltip: LH["Help"],
          status_bar_text: LH["Open the online user guide"]
      },
      "license" => {
          name: LH["License.."],
          tooltip: LH["Show licensing information"],
          status_bar_text: LH["Show licensing information"]
      },
      "pathmode" => {
          name: LH["Path Mode"],
          tooltip: LH["Path Mode"],
          status_bar_text: LH["Convert selected Profile Members to Path Mode"]
      },
      "revert_pathmode" => {
          name: LH["Revert Path Mode"],
          tooltip: LH["Revert Path Mode"],
          status_bar_text: LH["Revert selected paths back to Profile Members"]
      },
      "remove_trim" => {
          name: LH["Remove Trims"],
          tooltip: LH["Remove Trims"],
          status_bar_text: LH["Remove trims from the selected Profile Members"]
      }
  }

  def self.get_cursor(name, hotx, hoty)
    ext = ON_MAC ? ".pdf" : ".svg"
    icon = File.join(PATH, "cursors", "#{name}#{ext}")
    UI.create_cursor(icon, hotx, hoty)
  end

  def self.check_license
    lic = Licensing::License.new
    state = lic.get_state
    if state == Licensing::LICENSED || state == Licensing::TRIAL
      return true
    elsif state == Licensing::TRIAL_EXPIRED
      self.trial_expired
      return false
    else
      self.invalid_license_found
      return false
    end
  end

  # Action to be taken if an trial license is expired
  def self.trial_expired
    Sketchup.active_model.abort_operation # Should be safe to call this even if there is no current operation
    UI.messagebox(TRIAL_EXPIRED_MESSAGE)
    Sketchup.active_model.select_tool(nil)
    Licensing::LicenseInfo.instance.show
  end

  # Action to be taken if an invalid license is found
  # @param [InvalidLicenseException] err sub-class of RuntimeError
  def self.invalid_license_found(err = nil)
    Sketchup.active_model.abort_operation # Should be safe to call this even if there is no current operation
    UI.messagebox(INVALID_LICENSE_MESSAGE)
    Sketchup.active_model.select_tool(nil)
    Licensing::LicenseInfo.instance.show
  end

# @param [String] name, the name of the icon file (without extension) also the key to string hashes
# @param [callback code] block, the code to be executed when the cmd is activated
  def self.create_cmd(name, &block)
    ext = ON_MAC ? ".pdf" : ".svg"
    icon = File.join(PATH, "html/images/#{name}#{ext}")
    cmd_strings = CMD_HASH[name]
    cmd = UI::Command.new(cmd_strings[:name]) {  AutoUpdate.check_for_update(auto = true) {  block.call if (block) } }
    cmd.tooltip = cmd_strings[:tooltip]
    cmd.status_bar_text = cmd_strings[:status_bar_text]
    cmd.small_icon = icon
    cmd.large_icon = icon
    return cmd
  end

#adds a cmd to the menu and tb (if they are not nil)
#the name should be the icon base name
# @param [menu] menu, the menu to add the cmd to (nil if not to be added to a menu)
# @param [toolbar] tb, the toolbar to add the cmd to (nil if not to be added to a toolbar)
# @param [String] name, the name of the icon file (without extension) also the key to string hashes
# @param [callback code] block, the code to be executed when the cmd is activated
  def self.add_ui_cmd(menu, tb, name, &block)
    cmd = self.create_cmd(name) { block.call if (block) }
    tb.add_item(cmd) unless (tb.nil?)
    menu.add_item(cmd) unless (tb.nil?)
  end

####
  def self::build_ui
    main_menu = UI.menu("Plugins")
    main_menu.add_separator

    main_menu = main_menu.add_submenu("#{PLUGIN_NAME}")
    tb = UI::Toolbar.new("#{PLUGIN_NAME}")
    path = PATH

    self.add_ui_cmd(main_menu, tb, 'profile') { ProfileBuilderTool.create.show }
    self.add_ui_cmd(main_menu, tb, 'assembly') { AssemblyDialog.create.show }
    self.add_ui_cmd(main_menu, tb, 'path_select') { Sketchup.active_model.select_tool(PathSelectTool.new) }

    main_menu.add_separator
    tb.add_separator

    self.add_ui_cmd(main_menu, tb, 'extend') { Sketchup.active_model.select_tool(ExtendTool.new) }
    self.add_ui_cmd(main_menu, tb, 'hole_tool') { CutOpeningSettings.instance.show; }
    self.add_ui_cmd(main_menu, tb, 'trimsolid') { Sketchup.active_model.select_tool(TrimTool.new) }
    self.add_ui_cmd(main_menu, tb, 'trimface') { Sketchup.active_model.select_tool(TrimtoFaceTool.new) }
    self.add_ui_cmd(main_menu, tb, 'remove_trim')  { self.untrim_members }
    self.add_ui_cmd(main_menu, tb, 'edit_profile') { self.edit_profile }
    self.add_ui_cmd(main_menu, tb, 'edit_path') { self.edit_path }

    main_menu.add_separator
    tb.add_separator

    self.add_ui_cmd(main_menu, tb, 'pathmode') { PLUGIN.path_mode }
    self.add_ui_cmd(main_menu, tb, 'revert_pathmode') { PLUGIN.revert_path_mode }

    main_menu.add_separator
    # Commands below are only accessible via menu
    cmd = self.create_cmd('assign_mark_numbers') { MarkNumbers.instance.show }
    main_menu.add_item(cmd)
    cmd = self.create_cmd('import_plb') { PLUGIN.import_plb }
    main_menu.add_item(cmd)
    cmd = self.create_cmd('preferences') { Preferences.instance.show }
    main_menu.add_item(cmd)

    main_menu.add_separator

    cmd = self.create_cmd('help') { UI.openURL(HELP_URL) }
    main_menu.add_item(cmd)
    cmd = self.create_cmd('license') { Licensing::LicenseInfo.instance.show }
    main_menu.add_item(cmd)

    cmd = UI::Command.new(LH["Check for Update.."]) { AutoUpdate.check_for_update(auto = false) }
    cmd.tooltip = cmd.status_bar_text = LH["Open Update Dialog"]
    main_menu.add_item(cmd)

    #show toolbar on first load then remember last state
    state = tb.get_last_state
    case state
      when 1, -1
        tb.show
      when 0
        tb.hide
    end

    UI.add_context_menu_handler{ |menu|
      menu.add_separator
      menu = menu.add_submenu(PLUGIN_NAME)
      menu.add_item(LH["Rebuild Selected"]) { PLUGIN.rebuild_selected(Sketchup.active_model.selection.to_a) }
      menu.add_item(LH["Reverse Selected"]) { PLUGIN.reverse_selected(Sketchup.active_model.selection.to_a) }
    }
    PLUGIN.set_toolbar_to_update(tb) #so that we can refresh toolbar on mac
  end #build_ui

  unless file_loaded?(__FILE__)
    PLUGIN::build_ui if @did_load_extensions
    file_loaded(__FILE__)
  end

  def self.untrim_members
    model = Sketchup.active_model
    model.start_operation(LH["Untrim Members"], true)
    sel = Sketchup.active_model.selection
    pm_instances = sel.find_all { |e| PBAPI::Classify.profile_member?(e) }
    did_remove_trim = false
    pm_instances.each { |inst|
      pm = PBAPI::ProfileMember.new(inst)
      if pm.is_trimmed?
        pm.untrim(true)
        pm.untrim(false)
        pm.draw
        did_remove_trim = true
      end
    }
    unless did_remove_trim
      model.abort_operation
      message = LH["Please select a Profile Member that has been trimmed."]
      UI.messagebox(message)
    else
      model.commit_operation
    end
  rescue InvalidLicenseException
    PLUGIN.invalid_license_found
  end

##
  def PLUGIN.force_file_extension(file, ext)
    path = File.dirname(file)
    f_name = File.basename(file, ".*")
    f_name += ext
    path = File.join(path, f_name)
    return path
  end

##
  def self.edit_path
    model = Sketchup.active_model
    member = nil
    e = nil
    if !model.nil? && !model.active_path.nil?
      e = model.active_path.last
      if e && PBAPI::Classify.profile_member?(e)
        member = PBAPI::ProfileMember.new(e)
        if member.revolved?
          UI.messagebox(LH["You cannot edit the path of a revolved Profile Member."])
          return
        end
      elsif e && PBAPI::Classify.assembly?(e)
        member = PBAPI::Assembly.get(e)
      end
    end
    if e && member
      model.start_operation(LH["Start Path Edit"], true)
      member.start_path_edit
      e.add_observer(PLUGIN.path_edit_observer)
      model.commit_operation
      show_path_edit_instructions
    else
     UI.messagebox(LH["First, double-click a Profile Member or Assembly."])
    end
  end

  def self.show_path_edit_instructions
    has_shown = Sketchup.read_default(MAIN_DICTIONARY, SHOW_PATH_EDIT_TIP_KEY, false)
    unless has_shown
      message = LH["Use the SketchUp drawing tools to edit the path. Close the Group / Component when finished."]
      UI.messagebox(message)
      Sketchup.write_default(MAIN_DICTIONARY, SHOW_PATH_EDIT_TIP_KEY, true)
    end
  end

  def self.show_assembly_height_warning
    has_shown = Sketchup.read_default(MAIN_DICTIONARY, SHOW_HEIGHT_WARNING_KEY, false)
    unless has_shown
      message = LH["Assembly parts will be made unique, stretched, and offset to suit the overall height value. Sub-assembly spans will not be modified. Continue?"]
      result = UI.messagebox(message, MB_YESNO)
      Sketchup.write_default(MAIN_DICTIONARY, SHOW_HEIGHT_WARNING_KEY, true)
      if result == IDYES
        return true
      else
        return false
      end
    end
    true
  end

# Launches Profile Edit Mode by adding an observer to the Profile Member that the user is currently editing
  def self.edit_profile
    model = Sketchup.active_model
    if !model.nil? && !model.active_path.nil?
      e = model.active_path.last
      if PBAPI::Classify.profile_member?(e)
        pm = PBAPI::ProfileMember.new(e)
        model.start_operation(LH["Start Profile Edit"], true)
        profile_ents = pm.start_profile_edit
        self.setup_camera_from_profile_entities(profile_ents, pm)
        epmo = EditingProfileObserver.new(pm, model.active_view)
        epmo.ensure_profile_is_visible
        pm.instance.add_observer(epmo)
        model.commit_operation
        return
      end
    end
    UI.messagebox(LH["First, double-click a Profile Member."])
  end

  def self.setup_camera_from_profile_entities(profile_ents, pm)
    model = Sketchup.active_model
    cap_trans = pm.chain.point_transformation_at(0)
    offset = cap_trans.zaxis.transform(model.edit_transform)
    if profile_ents.is_a?(Sketchup::Face)
      bounds = profile_ents.bounds
    else
      raise "Unexpected value for Profile Entities" unless profile_ents.is_a?(Array)
      bounds = Geom::BoundingBox.new
      profile_ents.each { |edge| bounds.add(edge.bounds) }
    end
    target = bounds.center
    up = cap_trans.yaxis.transform(model.edit_transform)
    offset.length = bounds.diagonal.to_f * 2.0
    eye = target.offset(offset)
    fov = model.active_view.camera.fov
    perspective = true
    su_camera = Sketchup::Camera.new(eye, target, up, perspective, fov)
    model.active_view.camera = su_camera, EDIT_PROFILE_TRANSITION
  end

  def PLUGIN.parse_length(string)
    string.strip! #trialing space will make parse_length fail
    string.gsub!(/”|“/, '"') #curly quotes will make parse_length fail
    string.gsub!(/‘|’/, "'")
    string.gsub!(/-\s+/,'-')
    if string.match("'")
      string += '"' unless string[-1] == '"' #since we found ' assume user is dealing with feet and inches
    end
    return nil if Sketchup.parse_length(string) == nil
    Sketchup.parse_length(string) >= 0.0 ? Sketchup.parse_length(string) : -(Sketchup.parse_length(string.gsub('-', '')))
  end

  def PLUGIN.get_valid_filename(f_name)
    f_name = f_name.to_s
    illegal = "\\\/:*?\"<>|"
    return f_name.delete(illegal)
  end

  def self.path_mode
    sel = Sketchup.active_model.selection
    pm_groups = sel.find_all { |e| PBAPI::Classify.profile_member?(e) }
    if pm_groups.length > 0
      Sketchup.active_model.start_operation(LH["Path Mode"], true)
      pm_groups.each { |group|
        pm = PBAPI::ProfileMember.new(group)
        next if pm.revolved?
        pm.path_mode = PBAPI::PATH_MODE_NORMAL
        pm.draw
      }
      Sketchup.active_model.commit_operation
    else
      UI.messagebox(LH["First, select at least one Profile Member."])
    end
  end

  def self.revert_path_mode
    sel = Sketchup.active_model.selection
    pm_groups = sel.find_all { |e| PBAPI::Classify.profile_member?(e) }
    if pm_groups.length > 0
      Sketchup.active_model.start_operation(LH["Revert Path Mode"], true)
      pm_groups.each { |group|
        pm = PBAPI::ProfileMember.new(group)
        pm.path_mode = PBAPI::PATH_MODE_DISABLED
        pm.draw
      }
      Sketchup.active_model.commit_operation
    else
      UI.messagebox(LH["First, select at least one Profile Member in path mode."])
    end
  rescue InvalidLicenseException
    PLUGIN.invalid_license_found
  end

##
  def PLUGIN.import_plb
    default_path = File.dirname(get_last_profile_path)
    plb = UI.openpanel(LH["Import PLB"], default_path, "PLB|*.plb||")
    return unless plb
    new_library = ProfileLibrary.new(plb)
    library_name = File.basename(plb, ".plb")

    lib_dir = File.dirname(plb)
    new_lib_dir = UI.select_directory(
        title: LH["Select Output Folder"],
        directory: lib_dir
    )
    return if new_lib_dir.nil?

    profiles = new_library.profiles
    num_profiles = profiles.length
    count = 0
    pbar = ProgressBar.new(LH["Import PLB"], profiles) { |profile|
      profile.name = PLUGIN.get_valid_filename(profile.name)
      profile.save_skp(new_lib_dir)
    }

    Sketchup.active_model.active_view.animation = pbar
  end


  def PLUGIN.get_last_profile_path
    begin
      default_path = ""
      last_path = Sketchup.read_default("ProfileBuilder", "LastProfilePath", default_path)
    rescue
      PLUGIN.save_last_profile_path("")
      last_path = ""
    end

    return last_path
  end

  def PLUGIN.get_last_assembly_path
    begin
      default_path = ""
      last_path = Sketchup.read_default("ProfileBuilder", "LastAssemblyPath", default_path)
    rescue
      PLUGIN.save_last_assembly_path("")
      last_path = ""
    end
    return last_path
  end

#
  def PLUGIN.save_last_profile_path(path)
    path = File.absolute_path(path) #todo need to test on mac
    path.gsub!(/\\/,'\\\\\\\\') unless ON_MAC
    path.gsub!(/\"/,'\\\"') if ON_MAC
    last_path = Sketchup.write_default("ProfileBuilder", "LastProfilePath", path)
    PLUGIN.add_to_recent(path, REG_RECENT_PROFILES) #add it to the recent profiles list
    return last_path
  end

  def PLUGIN.save_last_assembly_path(path)
    path = File.absolute_path(path) #todo need to test on mac
    path.gsub!(/\\/,'\\\\\\\\') unless ON_MAC
    path.gsub!(/\"/,'\\\"') if ON_MAC
    last_path = Sketchup.write_default("ProfileBuilder", "LastAssemblyPath", path)
    PLUGIN.add_to_recent(path, REG_RECENT_ASSEMBLIES) #add it to the recent assembly list
    return last_path
  end

  def PLUGIN.add_to_recent(path, dictionary_key)
    if File.file?(path)
      path_array = Sketchup.read_default(MAIN_REG_SECTION, dictionary_key, []) #get the current array of recent
      path_array.delete(path) #remove the path if it is already in the list
      path_array.reverse! #reverse the array (newest will now be last)
      path_array.push(path) #add this path to the array
      path_array.reverse! #reverse the array (newest will now be first)
      if path_array.length > RECENT_ARRAY_MAX_LENGTH #if the array is to large reduce it to the max
        path_array = path_array.take(RECENT_ARRAY_MAX_LENGTH)
      end
      Sketchup.write_default(MAIN_REG_SECTION, dictionary_key, path_array) #save the new array as the recent assemblies
    end
  end

  def self.reverse_selected(ents)
    pm_ents = ents.find_all { |e| PBAPI::Classify.profile_member?(e) }
    a_ents = ents.find_all { |e| PBAPI::Classify.assembly?(e) }
    if pm_ents.length > 0 || a_ents.length > 0
      Sketchup.active_model.start_operation(LH["Reverse Selected"], true)
      self.reverse_members(pm_ents)
      self.reverse_assemblies(a_ents)
      Sketchup.active_model.commit_operation
    else
      UI.messagebox(LH["First select at least one Profile Member or Assembly."])
    end
  rescue InvalidLicenseException
    PLUGIN.invalid_license_found
  end

  def self.reverse_members(pm_ents)
    pm_ents.each{ |e|
      PBAPI::ProfileMember.new(e).reverse
    }
  end

  def self.reverse_assemblies(a_ents)
    a_ents.each{ |e|
      PBAPI::Assembly.get(e).reverse
    }
  end

  def self.rebuild_selected(ents)
    pm_ents = ents.find_all { |e| PBAPI::Classify.profile_member?(e) }
    a_ents = ents.find_all { |e| PBAPI::Classify.assembly?(e) }
    if pm_ents.length > 0 || a_ents.length > 0
      Sketchup.active_model.start_operation(LH["Rebuild Selected"], true)
      self.rebuild_members(pm_ents)
      self.rebuild_assemblies(a_ents)
      Sketchup.active_model.commit_operation
    else
      UI.messagebox(LH["First select at least one Profile Member or Assembly."])
    end
  rescue InvalidLicenseException
    PLUGIN.invalid_license_found
  end

  def self.rebuild_members(pm_ents)
    pm_ents.each{ |e|
      pm = PBAPI::ProfileMember.new(e)
      pm.draw
    }
  end

  def self.rebuild_assemblies(a_ents)
    a_ents.each{ |e|
      assembly = PBAPI::Assembly.get(e)
      assembly.draw
    }
  end

  # Assigns mark numbers to Profile Members
  # @param [Entities or Selection] ents
  # @param [String] prefix will form the start of the new name of the object
  # @param [String] suffix will form the end of the new name of the object
  # @param [Integer or EmptyString] start_number a number that will be auto-incremented and placed between
  #  the prefix and suffix
  # @return [Integer] returns the number of ProfileMembers that were renamed
  def self.assign_mark_numbers(ents, prefix, suffix, start_number)
    Sketchup.active_model.start_operation(LH["Assign Mark Numbers"])
    finder = Finder.new(ents)
    finder.find_profile_members
    transformations = finder.transformations
    renamed = 0
    current_number = start_number.to_i
    definitions_pms = []
    transformations.each { |pm, trans|
      # If the array already has values, iterate through it to check if this pm
      # should be combined with others with the same definition
      found_match = false
      definitions_pms.each_index { |index|
        first_pm_defn = definitions_pms[index].first[0].definition
        first_pm_trans = definitions_pms[index].first[1]
        if pm.definition == first_pm_defn && pm.same_scale?(trans, first_pm_trans)
          definitions_pms[index] << [pm, trans] # combine this PMs with similar PMs (same defn and scale)
          found_match = true
          break
        end
      }
      unless found_match
        definitions_pms << [[pm, trans]]
      end
    }

    definitions_pms.each { |list_of_pms|
      # Get the first Profile Member in this grouping.  All PMs in this group have the same
      # definition and scale so they must also share the same profile, length, height, and width
      pm = list_of_pms.first[0]
      trans = list_of_pms.first[1]
      p_name = pm.profile.name
      if start_number
        new_name = prefix + current_number.to_s + suffix
      else
        new_name = prefix + suffix
      end
      if new_name.include?("$P")
        new_name = new_name.gsub(/\$P/, p_name)
      end
      if new_name.include?("$L")
        new_name = new_name.gsub(/\$L/, Sketchup.format_length(pm.length(trans)) )
      end
      # For width and height, use the identity transformation.  Otherwise, if the PM is scaled,
      # the width and height values might be surprising
      if new_name.include?("$H")
        new_name = new_name.gsub(/\$H/, Sketchup.format_length(pm.height(Geom::Transformation.new)) )
      end
      if new_name.include?("$W")
        new_name = new_name.gsub(/\$W/, Sketchup.format_length(pm.width(Geom::Transformation.new)) )
      end
      # Rename all of the instances of this Profile Member.  Note that
      # all Profile members in the list below should have the same definition and scale
      list_of_pms.each { |pm_trans|
        pm = pm_trans[0]
        inst = pm.instance
        inst.name = new_name
        renamed += 1
      }
      current_number += 1
    }
    Sketchup.active_model.commit_operation
    renamed
  end

  class ProfileBuilderTool
    include PBAPI::Geometry
    include WebDialogHelper

    SNAP_PIXEL_DISTANCE = 50.0

    JUNCTION_STYLE_BUTTONS = {
        "junctionStyleContinuous" => PBAPI::JUNCTION_STYLE_CONTINUOUS,
        "junctionStyleMiter" => PBAPI::JUNCTION_STYLE_MITER,
        "junctionStyleButt" => PBAPI::JUNCTION_STYLE_BUTT
    }

    private_class_method :new  #used to make singleton
    @profile_library = nil
    @profile_builder_tool = nil  #will store the one and only instance of this class
    @editing_profile_member = nil  #will store the profile member whose path is currently being edited
    @editing_path_group = nil  #will store the path group that is currently being used for editing
    @modal_dialog = nil
    @is_modal = false

    attr_reader :profile, :rotation, :mirror, :placement_point, :x_offset, :y_offset, :smooth_angle, :size_x, :size_y
    ####create a new instance of the tool or return the already created instance
    def ProfileBuilderTool.create
      @profile_builder_tool = new unless @profile_builder_tool
      return @profile_builder_tool
    end

    def ProfileBuilderTool.close
      unless @profile_builder_tool.nil?
        @profile_builder_tool.close
      end
    end

    def ProfileBuilderTool.create_modal
      @modal_dialog = new unless @modal_dialog
      @modal_dialog.is_modal = true
      return @modal_dialog
    end

    def ProfileBuilderTool.close_modal
      unless @modal_dialog.nil?
        @modal_dialog.close
        @modal_dialog = nil
      end
    end

    def is_modal=(val)
      @is_modal = val
    end

    def refresh_profile_dropdowns
      if Sketchup.active_model
        refresh_materials_list
        refresh_layers_list
      end
    end

    def start_or_append_build_operation
      if @operation_started
        Sketchup.active_model.start_operation(LH["Build"], true, false, true) #append to previous
      else
        Sketchup.active_model.start_operation(LH["Build"], true) #start the original build operation
        @operation_started = true
      end
    end

    #should be called whenever the window gains focus so the material list has all the current sketchup models materials
    def refresh_materials_list
      mats = Sketchup.active_model.materials
      material_names = mats.collect{ |mat| mat.name }
      material_list = material_names.sort
      if @materials_list != material_list
        refreshMaterialsList("materialName")
        setMaterialName
        @materials_list = material_list
      end
    end

    #should be called whenever the window gains focus so the layers list has all the current sketchup models layers
    def refresh_layers_list
      #check if the list needs to be refreshed
      layers = Sketchup.active_model.layers
      layer_names = layers.collect{ |layer| layer.name }
      layers_list = layer_names.sort
      if @layers_list != layers_list
        refreshLayersList("layerName")
        setLayerName
        @layers_list = layers_list
      end
    end

    def initialize
      @layers_list = []
      @materials_list = []
      @ip1 = nil
      @ip2 = nil
      @xdown = 0
      @ydown = 0
      @chain = PBAPI::Chain.new
      @members = []	#will store the profile memberd object created by the tool
      @profile = PBAPI::Profile.new(DEFAULT_PROFILE)
      @lock_aspect = true
      @library = Sketchup.active_model.get_attribute('ProfileBuilder', 'library', nil)
      @cursor = PLUGIN.get_cursor("build_cursor", 1, 1)
      last_profile = PLUGIN.get_last_profile_path
      begin
        @profile.load_from_skp(last_profile) if FileTest.exist?(last_profile)
      rescue
        #continue
      end
      @dlg = nil
    end

    def init_dialog
      pref_key = "PB3Dialog"
      window_settings = {
          :dialog_title    => LH["Profile Dialog"],
          :preferences_key => pref_key,
          :scrollable      => true,
          :resizable       => true,
          :width           => 328,
          :height          => ON_MAC ? 508 : 548,
          :left            => 150,
          :top             => 150
      }
      @dlg = UI::HtmlDialog.new(window_settings)
      @dlg.set_file(File.join(PATH, "html/dialog.html"))
      @dlg.set_on_closed do
        UI.stop_timer(@timer) if @timer
        clean_up
        if @is_modal
          ProfileBuilderTool.close_modal
        else
          ProfileBuilderTool.close
        end
      end
    end

    def add_action_callbacks
      add_common_callbacks
      @dlg.add_action_callback("port_key") { |_, keycode|
        keycode = keycode.to_i
        case keycode
        when VK_DELETE, VK_HOME, VK_END, COPY_MODIFIER_KEY
          self.onKeyDown(keycode, false, 0, Sketchup.active_model.active_view) #This tool doesn't use flags so it can just be set to 0
        end
      }

      @dlg.add_action_callback("editProfile") { |d, p|
        folder = Sketchup.temp_dir
        file_name = @profile.name
        puts file_name
        @profile.save_skp(folder, file_name)
        Sketchup.open_file(File.join(folder, file_name + ".skp"))
      }

      @dlg.add_action_callback('ValueChanged') { |d, id, value|
        value = clean_string(value)
        #the following code handles any events that occur in the webdialog

        case id
        when "okButton"
          @callback.call(@profile.clone)
          @dlg.close
        when "cancelButton"
          self.close
        when "lockAspect"
          if @lock_aspect
            @lock_aspect = false
          else
            @lock_aspect = true
          end
          refreshlockaspect
        when "drawProfileButton"
          drawProfile
        when "saveButton"
          save_profile
        when "profileBrowseButton"
          browseProfile
        when "profileName"
          @profile.name = value
        when "placementPoint"
          @profile.placement_point = value.to_i
          self.refresh_preview
        when "rotation"
          if value
            v_h = value.split(":")
            if v_h.length == 2
              value = (v_h[0].to_f) / (v_h[1].to_f)
              value = Math::atan(value).radians
            end
            @profile.rotation = value.to_f
            self.refresh_preview
          end
          setValue("rotation", @profile.rotation) #updates the text field to the current value being used
        when "xOffset"
          @profile.x_offset = PLUGIN.parse_length(value).to_f
          self.refresh_preview
          set_length("xOffset", @profile.x_offset) #update text field with properly formatted value being used
        when "yOffset"
          @profile.y_offset = PLUGIN.parse_length(value).to_f
          self.refresh_preview
          set_length("yOffset", @profile.y_offset) #update text field with properly formatted value
        when "width"
          new_width = PLUGIN.parse_length(value).to_f
          if new_width > 0.0
            if @lock_aspect
              current_aspect = (@profile.width) / (@profile.height)
              new_height = new_width / current_aspect
              set_length("height", new_height)
              @profile.y_scale = new_height / @profile.default_height
              @profile.x_scale = new_width / @profile.default_width
            else
              @profile.x_scale = new_width / @profile.default_width
            end
            validateResetScale
            self.refresh_preview
          end
          set_length("width", @profile.width)
        when "height"
          new_height = PLUGIN.parse_length(value).to_f
          if new_height > 0.0
            if @lock_aspect
              current_aspect = (@profile.width) / (@profile.height)
              new_width = current_aspect * new_height
              set_length("width", new_width)
              @profile.y_scale = new_height / @profile.default_height
              @profile.x_scale = new_width / @profile.default_width
            else
              @profile.y_scale = new_height / @profile.default_height
            end
            validateResetScale
            self.refresh_preview
          end
          set_length("height", @profile.height)
        when "resetScale"
          @profile.x_scale = 1.0
          @profile.y_scale = 1.0
          width = Sketchup.format_length(@profile.width)
          height = Sketchup.format_length(@profile.height)
          setValue("width", (width))
          setValue("height", (height))
          validateResetScale
          self.refresh_preview
        when "smoothAngle"
          @profile.smooth_angle = value.to_f
        when "mirror"
          @profile.mirror = value
          self.refresh_preview
        when "layerName"
          @profile.layer_name = value
        when "materialName"
          @profile.material_name = value
        when "junctionStyleContinuous", "junctionStyleMiter", "junctionStyleButt"
          @profile.junction_style = value.to_i
          refresh_junction_style_button_states
        when "extrudeMode"
          @profile.extrude_mode = value.to_i
        when "drawMemberButton"
          Sketchup.active_model.select_tool(self)
        when "createProfileButton"
          create_profile
        when "editMemberButton"
          sel = Sketchup.active_model.selection
          if sel.empty?
            UI.messagebox(LH["Please select a profile member."])
          else
            assemblies = []
            sel.each{ |e| assemblies.push(e) if PBAPI::Classify.assembly?(e) }
            if !assemblies.empty?
              result = UI.messagebox(LH["Assemblies were found in your selection. Would you like to convert them to Profile Members?"], MB_YESNO)
              if result == IDYES
                begin
                  Sketchup.active_model.start_operation(LH["Convert assemblies to profile member"], true)
                  assemblies.each{ |ent|
                    PLUGIN::Converter.assembly_to_profile(ent, @profile) if PBAPI::Classify.assembly?(ent) #need to ensure entity is still an assembly (components)
                  }
                  Sketchup.active_model.commit_operation
                rescue InvalidLicenseException
                  PLUGIN.invalid_license_found
                end
              end
            end
            # perform the previous operation on the remaining profiles if any
            props = PMPropertiesSelector.instance
            props.show(self, "edit")
          end
        when "editPathButton"
          e = Sketchup.active_model.active_path.last
          if PBAPI::Classify.profile_member?(e)
            edit_path(PBAPI::ProfileMember.new(e))
          else
            UI.messagebox(LH["Please select a profile member."])
          end
        when "createAlongPathButton"
          build_along_path
        when "addToFavouritesButton"
          if @profile
            self.add_to_favourites(@profile)
            refreshLibraryList
            setValue("libraryName", @library)
          end
        when "acquireButton"
          tool = ProfileMemberPickerTool.new { |pm| self.set_values(pm) }
          tool.dialog = self
          Sketchup.active_model.select_tool(tool)
        when "selectByButton"
          props = PMPropertiesSelector.instance
          props.show(self, "select")
        when "revolveButton"
          tool = LatheTool.new(@profile)
          tool.dialog = self
          Sketchup.active_model.select_tool(tool)
        end
      }
    end

    def on_build_tool_activated
      cmd = "activateButton('drawMemberButton');"
      @dlg.execute_script(cmd)
    end

    def on_build_tool_deactivated
      cmd = "deactivateButton('drawMemberButton');"
      @dlg.execute_script(cmd)
    end

    def on_picker_tool_activated
      cmd = "activateButton('acquireButton');"
      @dlg.execute_script(cmd)
    end

    def on_picker_tool_deactivated
      cmd = "deactivateButton('acquireButton');"
      @dlg.execute_script(cmd)
    end

    def on_lathe_tool_activated
      cmd = "activateButton('revolveButton');"
      @dlg.execute_script(cmd)
    end

    def on_lathe_tool_deactivated
      cmd = "deactivateButton('revolveButton');"
      @dlg.execute_script(cmd)
    end


    # Executes the Build Along Path Tool
    def build_along_path
      valid = validate_create_along_path
      return unless valid
      paths = PathSelectTool.get_selected_paths
      if paths.length > 0
        Sketchup.active_model.start_operation(LH["Build along path"], true)
        paths.each { |path|
          @members = []
          @chain = path
          if @profile.junction_style == PBAPI::JUNCTION_STYLE_CONTINUOUS
            place_member
          else
            pm = PBAPI::ProfileMember.add(Sketchup.active_model.active_entities, @chain, nil, PLUGIN::Preferences.build_as_component)
            pm.set_from_profile!(@profile)
            @members = pm.draw_discontinuous(@profile.junction_style)
          end
        }
        Sketchup.active_model.commit_operation
      end
    rescue InvalidLicenseException
      PLUGIN.invalid_license_found
    end

    def validate_create_along_path
      sel = Sketchup.active_model.selection
      if sel.empty?
        UI.messagebox(LH["Please select at least one edge or one Profile Member."])
        return false
      end
      if @profile == nil
        UI.messagebox(LH["Profile not loaded."])
        return false
      end
      return true
    end

    ##
    def get_selected_edges
      edges = Sketchup.active_model.selection.grep(Sketchup::Edge)
      return edges
    end

    ##
    def clean_up
      @browser.close if @browser
      PMPropertiesSelector.instance.close if PMPropertiesSelector.exists?
    end

    ##
    def browseProfile
      @browser.close if @browser
      last_path = PLUGIN.get_last_profile_path
      if FileTest.exist?(last_path)
        folder = File.dirname(last_path)
      else
        folder = File.join(File.dirname(__FILE__), "Libraries", "Profile Examples")
        if !FileTest.exist?(folder)
          folder = Dir.pwd
        end
      end
      @browser = SKPBrowser.new(folder, LH["Select Profile"], profiles = true) { |skp|
        status = true
        begin
          @profile.load_from_skp(skp)
        rescue IOError
          status = browse_profile_error(LH["Invalid SketchUp File"])
        rescue PBAPI::Profile::InvalidProfileSKP => err
          status = browse_profile_error(err.message)
        rescue PBAPI::FileVersionNewer
          status = browse_profile_error(LH["The file you are trying to open is not compatible with your version of SketchUp. Please request the creator of the file to re-save it as version: "] + "#{Sketchup.version.to_i}")
        rescue
          status = browse_profile_error(LH["An unknown error occurred when trying to load the selected file."])
          raise
        end
        if status
          @browser.close
          refresh_dialog
          PLUGIN.save_last_profile_path(skp)
          @dlg.bring_to_front
        end
      }
      @browser.show
    end

    #reduce code duplication for different error cases. returns false so status can be set
    def browse_profile_error(message)
      PLUGIN.save_last_profile_path("")
      UI.messagebox(message)
      return false
    end

    ##
    def create_profile
      sel = Sketchup.active_model.selection
      added = 0
      new_profiles = []
      found_face = false
      found_edge = false
      copy_profile = false
      default_name = LH["Profile"]

      if @profile && sel.empty?
        default_name = @profile.name + LH["_copy"]
        copy_profile = true
      end

      faces = sel.grep(Sketchup::Face)
      edges = sel.grep(Sketchup::Edge)

      found_face = true if faces.length > 0
      found_edge = true if edges.length > 0

      if found_face || found_edge || copy_profile
        skp = UI.inputbox([LH["Name"]], [default_name], LH["Create New Profile"])
        return unless skp
      else
        UI.messagebox(LH["Please select a face or polyline."])
        return
      end

      p_name = skp[0]

      if found_face && faces.length == 1
        p = PBAPI::Profile.new(faces.first)
        p.name = p_name
        @profile = p
        new_profiles.push(p)
      elsif found_face && faces.length > 1
        dir = UI.select_directory(title: LH["Choose Folder for Saving Multiple Profiles"])
        return if dir.nil?
        faces.each{ |e|
          if added >= 1
            name = p_name + "#{added}"
          else
            name = p_name
          end
          p = PBAPI::Profile.new(e)
          p.name = name
          new_profiles.push(p)
          p.save_skp(dir)
          added += 1
        }
        UI.messagebox(LH["Number of Profiles saved: "] + "#{dir}")
      elsif found_edge
        paths = PathSelectTool.get_selected_paths
        if paths.empty?
          UI.messagebox(LH["A valid path was not found."])
          return
        elsif paths.length > 1
          UI.messagebox(LH["Please select a single path."])
          return
        end
        path = paths.first
        plane = [Geom::Point3d.new(0, 0, 0), Geom::Vector3d.new(0, 0, 1)]
        off_plane = path.find{ |p| !p.on_plane?(plane) }
        if off_plane
          UI.messagebox(LH["All points on a polyline profile must be drawn on the ground plane."])
          return
        end
        chain = path.clone
        p = PBAPI::Profile.new(chain)
        p.dimension = 1
        p.name = p_name
        new_profiles.push(p)
        @profile = p
        added += 1
      elsif sel.empty?
        if @profile
          p = @profile.clone
          p.name = p_name
          new_profiles.push(p)
          added += 1
          @profile = p
        end
      end

      if new_profiles.first
        refresh_dialog
      else
        UI.messagebox(LH["Please select at least one face."])
      end
    end

    ##
    def drawProfile
      Sketchup.active_model.start_operation(LH["Stamp Profile"], true, true) #whatever operation comes after this one will be appended into one combined operation USE WITH CAUTION
      model = Sketchup.active_model
      defns = model.definitions
      profile_defn = defns.add(@profile.name)
      behavior = profile_defn.behavior
      behavior.is2d = true
      behavior.snapto = SnapTo_Arbitrary
      ents = profile_defn.entities
      @profile.draw(ents)
      Sketchup.active_model.commit_operation
      model.place_component(profile_defn, false) #this will create an operation when the component is placed. WARNING: if not placed it will not create an operation!!
    end

    #################################
    def show(&block)
      return unless PLUGIN.check_license
      show_dialog(&block)
      refresh_dropdowns_at_interval
    end

    # this method is used for editing profiles that are part of assemblies or used as holes.
    # The dialog will not actually be truly modal (blocking). It just called modal to indicate a special mode of the
    # dialog where the user needs to click OK or cancel in order to proceed.
    def show_modal(profile, &block)
      @profile = profile.clone unless (profile == nil)
      @callback = block # will be called when user clicks the OK button of the modal dialog version
      show_dialog
      refresh_dropdowns_at_interval
    end

    def refresh_dropdowns_at_interval
      UI.stop_timer(@timer) if @timer
      @timer = UI.start_timer(0.25, true) { refresh_profile_dropdowns }
    end

    ##
    def close
      @dlg.close
    end

    def visible?
      @dlg.visible?
    end

    #name should only be 'width' or 'height'
    def disabled_input_field_if_it_is_zero(name)
      disabled = false
      value = ('width' == name) ? @profile.width : @profile.height
      if value == 0
        cmd = "document.getElementById('#{name}').setAttribute('disabled', true);"
        @dlg.execute_script(cmd)
        disabled = true
      else
        cmd = "document.getElementById('#{name}').removeAttribute('disabled');"
        @dlg.execute_script(cmd)
        disabled = false
      end
      return disabled
    end

    #checks if the current profile has a default height or width of 0
    #if it does that 0 value field is disabled and the link is locked as broken since changing a 0 value field gives errors
    def disable_link_and_height_or_width_if_zero
      if @profile
        width_disabled = disabled_input_field_if_it_is_zero('width')
        height_disabled = disabled_input_field_if_it_is_zero('height')
        if width_disabled || height_disabled
          @lock_aspect = false
          refreshlockaspect
          cmd = "document.getElementById('lockAspect').className = '';"
          @dlg.execute_script(cmd)
          cmd = "document.getElementById('lockAspect').setAttribute('disabled', true);"
          @dlg.execute_script(cmd)
        else
          cmd = "document.getElementById('lockAspect').className = 'hover';"
          @dlg.execute_script(cmd)
          cmd = "document.getElementById('lockAspect').removeAttribute('disabled');"
          @dlg.execute_script(cmd)
        end
      end
    end

    ##
    def refresh_dialog
      refreshMaterialsList("materialName")
      refreshLayersList("layerName")
      getValuesFromSU
      refreshlockaspect
      disable_link_and_height_or_width_if_zero #disable height or width and the link button if single line profile
      set_modal_display if @is_modal
    end

    ##
    def refreshlockaspect
      if @lock_aspect
        @dlg.execute_script("lockAspect();")
      else
        @dlg.execute_script("unlocklockAspect();")
      end
    end

    ##
    def set_modal_display
      cmd = "showModalDisplay();"
      @dlg.execute_script(cmd)
    end

    def set_length(id, val)
      length = Sketchup.format_length(val)
      setValue(id, length)
    end

    #################################
    def getValuesFromSU
      if @profile
        setValue("profileName", @profile.name)
        set_length("width", @profile.width)
        set_length("height", @profile.height)

        set_length("xOffset", @profile.x_offset)
        set_length("yOffset", @profile.y_offset)

        setValue("placementPoint", @profile.placement_point)
        setValue("rotation", @profile.rotation)
        refresh_junction_style_button_states
        setValue("extrudeMode", @profile.extrude_mode)

        setMaterialName
        setLayerName

        if @profile.mirror
          setCheckbox("mirror", "true")
        else
          setCheckbox("mirror", "false")
        end

        @lock_aspect = true
        validateResetScale
        refresh_preview
      end
    end

    def refresh_junction_style_button_states
      JUNCTION_STYLE_BUTTONS.each { |id, val|
        if val == @profile.junction_style
          @dlg.execute_script("setButtonActive('#{id}')");
        else
          @dlg.execute_script("setButtonInactive('#{id}')");
        end
      }
    end

    ##
    def validateResetScale
      if @profile.scaled?
        cmd = "$('#resetScale').prop('disabled', false);"
      else
        cmd = "$('#resetScale').prop('disabled',true);"
      end
      @dlg.execute_script(cmd)
    end

    ##
    def setMaterialName
      id = "materialName"
      if @profile
        mat_names = Sketchup.active_model.materials.collect{ |mat| mat.name }
        if mat_names.index(@profile.material_name)
          setValue(id, (@profile.material_name))
        else
          @profile.material_name = "Default"
          setValue(id, (@profile.material_name))
        end
      end
    end

    ##
    def setLayerName
      id = "layerName"
      if @profile
        layer_names = Sketchup.active_model.layers.collect{ |layer| layer.name }
        if layer_names.index(@profile.layer_name)
          setValue(id, (@profile.layer_name))
        else
          @profile.layer_name = "Layer0"
          setValue(id, (@profile.layer_name))
        end
      end
    end

    ###
    def copy_to_profile
      profile = @profile.dup
      return profile
    end

    ############################
    def set_values(pm)
      @profile.set_from_profile_member(pm)
      self.refresh_dialog
      return
    end

    ##
    def save_profile
      begin
        default_name = @profile.name.strip
        default_name = LH["Profile"] if default_name == ""
        last_path = PLUGIN.get_last_profile_path
        if FileTest.exist?(last_path)
          folder = File.dirname(last_path)
        else
          folder = Dir.pwd
        end
        skp = UI.savepanel(LH["Save Profile"], folder, default_name + ".skp")
        return unless skp

        if skp =~ /DM_ProfileBuilder/
          UI.messagebox(LH["Profiles cannot be saved within the Profile Builder plugin folder.  Please choose another location."])
          return
        end

        f_name = File.basename(skp, ".*")
        f_name = PLUGIN.get_valid_filename(f_name)
        @profile.save_skp(File.dirname(skp), f_name)
        PLUGIN.save_last_profile_path(skp)
        UI.messagebox(LH["Profile Saved."])
      rescue
        UI.messagebox(LH["Unable to save Profile."])
        raise
      end
    end

    def refresh_preview
      preview_profile(
          @dlg,
          @profile,
          @profile.placement_point,
          @profile.rotation,
          @profile.x_offset,
          @profile.y_offset,
          @profile.mirror,
          @profile.width,
          @profile.height,
          "Canvas")
    end

    ################################
    def preview_profile(dlg, profile, placement_point = 5, rotation = 0.0, x_offset = 0.0, y_offset = 0.0, mirror = false, x_size = nil, y_size = nil, canvis_id = nil)
      canvas = 128 #the size of the web dialog canvas
      cent = (canvas.to_f / 2.0).to_i
      center = Geom::Point3d.new(cent, cent, 0)

      profile_trans = profile.get_2d_orientation_transformation
      pts = profile.points.collect{ |p| p.transform(profile_trans) } #shift points so (0,0,0) is the specified placement point
      pts.reverse! if profile.mirror?
      preview_origin = Geom::Point3d.new(0, 0, 0) #this point should always be (0, 0, 0) no matter what placement option is set

      box = Geom::BoundingBox.new
      box.add(pts)
      box.add(preview_origin)
      vec = center - box.center
      trans1 = Geom::Transformation.translation(vec)

      new_points = pts.collect{ |p| p.transform(trans1) } #shift shape so it is in the center of the canvas
      preview_origin = preview_origin.transform(trans1)  #shift origin as well

      box = Geom::BoundingBox.new
      box.add(new_points)
      box.add(preview_origin) #line shouldn't be needed leaving to ensure transformation are working
      xmax = box.max.x - box.min.x
      ymax = box.max.y - box.min.y
      xymax = [xmax, ymax].max.to_f
      scale = (canvas.to_f / xymax)
      trans2 = Geom::Transformation.scaling(center, scale, -scale, scale)
      new_points = new_points.collect{ |p| p.transform(trans2) } #scale shape so it fills the canvas
      preview_origin = preview_origin.transform(trans2) #scale the origin as well

      xpoints = new_points.collect{ |p| p.x.to_i }
      ypoints = new_points.collect{ |p| p.y.to_i }
      x_array = "new Array("
      y_array = "new Array("
      x_array += xpoints.join(",")
      y_array += ypoints.join(",")
      x_array += ")"
      y_array += ")"

      cmd = "clearCanvas()"
      dlg.execute_script(cmd)
      cmd = "drawProfile(#{x_array},#{y_array},\"#{canvis_id}\")" if profile.is_2d?
      cmd = "drawPolyline(#{x_array},#{y_array},\"#{canvis_id}\")" if profile.is_1d?
      dlg.execute_script(cmd)

      #draw the holes in the profile
      profile.holes.each{ |hole|
        #new code for rotation,mirror, and offset
        new_hole = hole.collect{ |p| p.transform(profile_trans) }
        new_hole = new_hole.collect{ |p| p.transform(trans1) }
        new_hole = new_hole.collect{ |p| p.transform(trans2) }

        xpoints = new_hole.collect{ |p| p.x.to_i }
        ypoints = new_hole.collect{ |p| p.y.to_i }
        x_array = "new Array("
        y_array = "new Array("
        x_array += xpoints.join(",")
        y_array += ypoints.join(",")
        x_array += ")"
        y_array += ")"
        cmd = "drawHole(#{x_array},#{y_array},\"#{canvis_id}\")"
        dlg.execute_script(cmd)
      }

      origin_dot_size = 6.0
      x = preview_origin.x.to_i - (origin_dot_size / 2.0)
      y = preview_origin.y.to_i - (origin_dot_size / 2.0)
      cmd = "drawOrigin(#{x},#{y},\"#{canvis_id}\")"
      dlg.execute_script(cmd)
    end

    ##################################################
    def activate
      on_build_tool_activated if @dlg
      @ip1 = Sketchup::InputPoint.new
      @ip2 = Sketchup::InputPoint.new
      @ip = Sketchup::InputPoint.new
      @last_point = Geom::Point3d.new(0, 0, 0)
      @drawn = false
      @members = []
      @building = false
      @operation_started = false

      # This sets the label for the VCB
      Sketchup::set_status_text(LH["Length"], SB_VCB_LABEL)
      Sketchup::set_status_text(nil, SB_VCB_VALUE)
      self.reset(nil)
    end

    ###################################################
    def deactivate(view)
      on_build_tool_deactivated if @dlg
      view.invalidate if @drawn
      view.lock_inference
      @members = []
    end

    #######################
    def onSetCursor
      UI.set_cursor(@cursor) if @cursor
    end

    ###################################################
    #method to tell SU the extent of what to draw
    def getExtents
      box = Sketchup.active_model.bounds
      box.add(@ip1.position) if @ip1.valid?
      box.add(@ip.position) if @ip.valid?
      box.add(@ip2.position) if @ip2.valid?

      return box
    end

    def resume(view)
      view.invalidate
    end

    ##
    def get_status_text
      case @state
        when 0
          text = LH["Click start point"]
        else
          text = LH["Click next point or enter distance. ESC, ENTER, or RIGHT-CLICK to finish. BACKSPACE to undo last point."]
      end
      return text
    end

    ####################################################
    def onMouseMove(flags, x, y, view)
      @x_current = x
      @y_current = y

      if @state == 0
        Sketchup::set_status_text(get_status_text)
        @ip.pick(view, x, y)
        if @ip != @ip1
          view.invalidate if @ip.display? || @ip1.display?
          @ip1.copy!(@ip)
          view.tooltip = @ip1.tooltip
        end
      else
        @dlg.execute_script("ensure_tool_settings_are_updated();") #make sure all the settings are updated

        Sketchup::set_status_text(get_status_text)
        @ip2.pick(view, x, y, @ip1)
        if should_snap_to_start?(view)
          @ip2 = Sketchup::InputPoint.new(@chain.first)
        end
        view.tooltip = @ip2.tooltip if @ip2.valid?
        view.invalidate

        length = @ip1.position.distance(@ip2.position)
        Sketchup::set_status_text length.to_s, SB_VCB_VALUE
      end
    end

    def should_snap_to_start?(view)
      @chain.num_points > 2 && screen_distance(view, @ip2.position, @chain.first) < SNAP_PIXEL_DISTANCE
    end

    def lock_axis(view, axis)
      if view.inference_locked?
        view.lock_inference
      end
      if @state == 1 && @ip2.valid?
        p2 = @ip2.position
        p1 = @ip1.position
        vec = p1.vector_to(p2)
        len = vec.dot(axis)
        len = 1.0 if len == 0.0
        offset_vector = Geom::Vector3d.new(axis)
        offset_vector.length = len
        @ip2 = Sketchup::InputPoint.new(@ip1.position.offset(offset_vector))
        view.lock_inference(@ip1, @ip2)
      end
    end

    ###################################################
    def onLButtonDown(flags, x, y, view)
      if @state == 0
        @ip1.pick(view, x, y)
        if @ip1.valid?
          @state = 1
          @last_point = @ip1.position
          @chain.add(@last_point)
          Sketchup::set_status_text(get_status_text)
          @xdown = x
          @ydown = y
        end
      else
        if @ip2.valid?
          start_or_append_build_operation
          @chain.add(@ip2.position)
          @last_point = @chain.points.last
          place_member
          @ip1 = Sketchup::InputPoint.new(@last_point)
          @drawn = false
          @left_down = @right_down = @up_down = @down_down = false
          if @chain.num_points > 2 && @chain.closed_path?
            self.reset(view)
          end
          Sketchup.active_model.commit_operation
        end
      end

      # Clear any inference lock
      view.lock_inference
      view.invalidate
    rescue InvalidLicenseException
      PLUGIN.invalid_license_found
    end

    ##
    def onReturn(view)
      self.reset(view)
    end

    def close_path
      start_or_append_build_operation
      @chain.add(@chain.first)
      place_member
      self.reset(Sketchup.active_model.active_view)
      Sketchup.active_model.commit_operation
    rescue InvalidLicenseException
      PLUGIN.invalid_license_found
    end

    ##
    def getMenu(menu)
      if @chain.num_points > 2
        menu.add_item(LH["Close the Path"]) { close_path }
      end
      if @chain.num_points > 1
        menu.add_item(LH["Finish"]) { self.reset(Sketchup.active_model.active_view) }
      end
    end

    #########################
    def onKeyDown(key, repeat, flags, view)
      case key
        when VK_END
          @profile.rotation += 45.0
          if @profile.rotation > 360.0
            @profile.rotation = @profile.rotation % 360.0
          end
          getValuesFromSU
        when VK_HOME
          @profile.placement_point += 1
          if @profile.placement_point == 10
            @profile.placement_point = 1
          end
          getValuesFromSU
        when CONSTRAIN_MODIFIER_KEY
          @shift_down_time = Time.now
          # if we already have an inference lock, then unlock it
          if view.inference_locked?
            # calling lock_inference with no arguments actually unlocks
            view.lock_inference
          elsif @state == 0 && @ip1.valid?
            view.lock_inference(@ip1)
          elsif @state == 1 && @ip2.valid?
            view.lock_inference(@ip2, @ip1)
          end
        when VK_LEFT
          @left_down =! @left_down
          if @left_down
            lock_axis(view, Y_AXIS)
          else
            view.lock_inference if view.inference_locked?
          end
          @right_down = @up_down = @down_down = false
        when VK_RIGHT
          @right_down =! @right_down
          if @right_down
            lock_axis(view, X_AXIS)
          else
            view.lock_inference if view.inference_locked?
          end
          @left_down = @down_down = @up_down = false
        when VK_DOWN
          @down_down =! @down_down
          if @down_down
            lock_axis(view, Z_AXIS)
          else
            view.lock_inference if view.inference_locked?
          end
          @up_down = @left_down = @right_down = false
        when VK_UP
          @up_down =! @up_down
          if @up_down
            lock_axis(view, Z_AXIS)
          else
            view.lock_inference if view.inference_locked?
          end
          @down_down = @left_down = @right_down = false
      end
      view.invalidate
    rescue InvalidLicenseException
      PLUGIN.invalid_license_found
    end

    #################
    def onKeyUp(key, repeat, flags, view)
      if key == CONSTRAIN_MODIFIER_KEY &&
          view.inference_locked? &&
          (Time.now - @shift_down_time) > 0.5
        view.lock_inference
      end
      if key == VK_DELETE || key == 8 || key == 127
        if @chain.num_points > 1
          @chain.pop
          @last_point = @chain.last
          @ip1 = Sketchup::InputPoint.new(@last_point)
          start_or_append_build_operation
          if @chain.num_points > 1 && @profile.junction_style == PBAPI::JUNCTION_STYLE_CONTINUOUS
            place_member
          else
            delete_member(view)
          end
          Sketchup.active_model.commit_operation
        end
      end
      view.invalidate
    end

    ##
    def onUserText(text, view)
      # The user may type in something that we can't parse as a length
      # so we set up some exception handling to trap that
      begin
        value = text.to_l
      rescue
        # Error parsing the text
        UI.beep
        puts "Cannot convert #{text} to a Length"
        value = nil
        Sketchup::set_status_text("", SB_VCB_VALUE)
      end
      return if !value

      # Compute the direction and the second point
      pt1 = @last_point
      vec = @ip2.position - pt1
      if vec.length == 0.0
        return
      end
      vec.length = value
      pt2 = pt1 + vec
      @last_point = pt2
      @chain.add(@last_point)

      # Create a line
      start_or_append_build_operation
      place_member
      ip_temp = Sketchup::InputPoint.new(pt2)   #Create a new inputpoint and set it to the point defined by the input length
      @ip1.copy!(ip_temp)
      @drawn = false
      Sketchup.active_model.commit_operation
    rescue InvalidLicenseException
      PLUGIN.invalid_license_found
    end

    ##
    def draw(view)
      if @ip1.valid?
        if @ip1.display?
          @ip1.draw(view)
          @drawn = true
        end

        if @ip2.valid?
          @ip2.draw(view) if @ip2.display?
          view.set_color_from_line(@last_point, @ip2)
          self.draw_geometry(@last_point, @ip2.position, view)
          @drawn = true
        end
      end
    rescue InvalidLicenseException
      PLUGIN.invalid_license_found
    end

    # onCancel is called when the user hits the escape key
    def onCancel(flag, view)
      self.reset(view)
    end


    # The following methods are not directly called from SketchUp.  They are
    # internal methods that are used to support the other methods in this class.

    # Reset the tool back to its initial state
    def reset(view)
      # This variable keeps track of which point we are currently getting
      @state = 0

      # Display a prompt on the status bar
      Sketchup::set_status_text(get_status_text)

      # clear the InputPoints
      @ip1.clear
      @ip2.clear

      if view
        view.tooltip = nil
        view.invalidate if @drawn
      end

      @members = []
      @chain.clear
      @drawn = false
      @left_down = false
      @right_down = false
      @down_down = false
      @up_down = false
      view.lock_inference if view && view.inference_locked?
      if @operation_started
        @operation_started = false
      end
    end

    ##
    def delete_member(view)
      if @members.last && @members.last.instance.valid?
        model = view.model
        ents = model.active_entities
        ents.erase_entities([@members.last.instance])
        @members.pop
      end
    end

      ####
      def place_member
        entities = Sketchup.active_model.active_entities
        @chain.axis_mode = @profile.extrude_mode
        case @profile.junction_style
          when PBAPI::JUNCTION_STYLE_BUTT, PBAPI::JUNCTION_STYLE_MITER
            # For discontinuous build modes, create a new PM from the last two points of the tool chain
            sub_chain = PBAPI::Chain.new(@chain[-2..-1])
            first_trans = PBAPI::ProfileMember.default_transformation(@chain, entities)
            # TODO: I tried to change the code below to use the new ProfileMember.discontinuous_trans method
            #  which was added in version 3.0.5.  However, when building a butt-joined PM using the Build Tool,
            #  the individual transformations of the butt-joined PMs were not being calculated correctly.  Therefore,
            #  we will keep this code below as-is for now even though there is duplication between the two.  Try
            #  to fix this in the future.
            if @chain.num_points == 2
              # If there are only two points in the tool chain (i.e. first PM to be created)
              # set the transformation to be that of the default transformation (takes into account drawing axes)
              trans = first_trans
            else
              # Get the chain transformation in terms of world coordinates
              temp_chain = @chain.clone
              # Get the world chain in terms of the first PM transformation
              temp_chain.transform!(first_trans.inverse)
              # Get the last edge transformation (in terms of the first PM transformation)
              trans = temp_chain.transformation_at(temp_chain.num_points - 2)
              # Transform this back in terms of world coordinates to get the transformation of this PM
              trans = first_trans * trans
              trans = Geom::Transformation.new(trans.xaxis, trans.zaxis.reverse, trans.yaxis, trans.origin)
            end
            pm = PBAPI::ProfileMember.add(entities, sub_chain, trans, PLUGIN::Preferences.build_as_component)
          else #PBAPI::JUNCTION_STYLE_CONTINUOUS
            if @members.last && @members.last.instance.valid?
              pm = @members.pop
              pm.set_chain(@chain, pm.transformation)
            else
              pm = PBAPI::ProfileMember.add(entities, @chain, nil, PLUGIN::Preferences.build_as_component)
            end
        end
        pm.set_from_profile!(@profile)
        pm.draw
        PBAPI.apply_butt_joint(pm, @members.last) if @profile.junction_style == PBAPI::JUNCTION_STYLE_BUTT
        PBAPI.apply_miter_joint(pm, @members.last) if @profile.junction_style == PBAPI::JUNCTION_STYLE_MITER
        @members.push(pm)

        # If the user just completed a closed path, we need to apply a joint between the first pm and the last one
        if @chain.closed_path?
          PBAPI.apply_butt_joint(@members.first, pm) if @profile.junction_style == PBAPI::JUNCTION_STYLE_BUTT
          PBAPI.apply_miter_joint(@members.first, pm) if @profile.junction_style == PBAPI::JUNCTION_STYLE_MITER
        end
      end

      ###############################################
      # Draw the temporary geometry
      def draw_geometry(pt1, pt2, view)
        view.line_width = 3
        view.draw_line(pt1, pt2)

        vec = pt2 - pt1
        if vec.length > 0
          previewer = PBAPI::PMPreviewer.new(view)
          previewer.profile = @profile
          temp_chain = PBAPI::Chain.new
          temp_chain.set_path(@chain.points)
          temp_chain.add(pt2)
          pm_trans = PBAPI::ProfileMember.default_transformation(temp_chain, Sketchup.active_model.active_entities)
          previewer.transformation = pm_trans
          previewer.chain = temp_chain
          previewer.draw(last_edge_only = true)
        end
      end
    end

#copy of ProfileBuilderTool used by the hole tool
  class ProfileBuilderToolHole < ProfileBuilderTool
  end

################################
  class PMPropertiesSelector   #class to show the dialog for selecting profile member properties to edit or select by

    include WebDialogHelper
    extend Singleton

    def initialize
      @tool = nil
      @mode =  nil
      @results = []
      @dlg = nil
    end

    def init_dialog
      window_settings = {
          :dialog_title    => LH["Select Properties"],
          :preferences_key => "PB3_Prop_Sel",
          :scrollable      => false,
          :resizable       => true,
          :width           => 448,
          :height          => ON_MAC ? 208 : 228,
          :left            => 387,
          :top             => 172
      }
      @dlg = UI::HtmlDialog.new(window_settings)
      @dlg.set_file(File.join(PATH, "html/edit.html"))
    end

    def add_action_callbacks
      add_common_callbacks
      @dlg.add_action_callback("apply_settings") { |d, p|
        model = Sketchup.active_model
        params = p.split("|")
        profile = params[0] == "true"
        rotation = params[1] == "true"
        mirror = params[2] == "true"
        pp = params[3] == "true"
        x_offset = params[4] == "true"
        y_offset = params[5] == "true"
        material = params[6] == "true"
        layer = params[7] == "true"
        junctions = params[8] == "true"
        extrude_mode = params[9] == "true"
        @results = [profile, rotation, mirror, pp, x_offset, y_offset, material, layer, junctions, extrude_mode]
        @dlg.close
        if @mode == "edit"
          model.start_operation(LH["Edit Member"], true)
          model.set_attribute('ProfileBuilder', 'PMPSValues', p)
          self.edit_member if (@mode == "edit")
          model.commit_operation
        elsif @mode == "select"
          model.start_operation(LH["Store Properties"], true)
          model.set_attribute('ProfileBuilder', 'PMPSValues', p)
          model.commit_operation
          self.select_by
        end
      }
      @dlg.add_action_callback("cancel") { |d, p| @dlg.close }
    end

    def show(tool, mode)
      if @dlg && @dlg.visible?
        # do nothing
      else
        # Set the mode of the dialog (and button label) only if it is not currently visible
        @tool = tool  # reference to the ProfileBuilder tool object that called this dialog
        @mode = mode  # mode must be a string either "edit" or "select"  - this is a test due to asynchronous behaviour of webdialogs on Mac systems
      end
      show_dialog
    end

    def close
      @dlg.close
    end

    ##############
    def edit_member
      model = Sketchup.active_model
      sel = model.selection

      if !@results.empty?
        edited_members = []
        sel.to_a.each{ |e|
          if PBAPI::Classify.profile_member?(e)
            edited_members.concat(modify_member_settings_and_redraw(e))
          end
        }
        if edited_members.length == 0
          UI.messagebox(LH["No Profile Members were selected."])
        end
        sel.clear
        sel.add(edited_members.collect { |pm| pm.instance })
      end
    rescue InvalidLicenseException
      PLUGIN.invalid_license_found
    end

    #creates a new pm and settings based on edit member window
    def modify_member_settings_and_redraw(e)
      tool_profile = @tool.copy_to_profile
      if e.class == Sketchup::Group
        e.make_unique
      end
      pm = PBAPI::ProfileMember.new(e)
      if @results[0]
        pm.profile = tool_profile
        e.name = tool_profile.name
        pm.x_scale = tool_profile.x_scale #the stored width must be overwritten if the profile changes
        pm.y_scale = tool_profile.y_scale #the stored height must be overwritten if the profile changes
      end
      pm.rotation = tool_profile.rotation if @results[1]
      pm.mirror = tool_profile.mirror if @results[2]
      pm.placement_point = tool_profile.placement_point if @results[3]
      pm.x_offset = tool_profile.x_offset if @results[4]
      pm.y_offset = tool_profile.y_offset if @results[5]
      pm.material_name = tool_profile.material_name if @results[6]
      pm.layer = tool_profile.layer_name if @results[7]
      pm.extrude_mode = tool_profile.extrude_mode if @results[9]
      draw_discontinuous = false
      if @results[8]
        old_style = pm.junction_style
        pm.junction_style = tool_profile.junction_style
        if old_style == PBAPI::JUNCTION_STYLE_CONTINUOUS && old_style != tool_profile.junction_style
          draw_discontinuous = true
        end
      end
      if draw_discontinuous
        members = pm.draw_discontinuous(pm.junction_style)
        pm.erase!
      else
        pm.draw
        members = [pm]
      end
      members
    end

    def select_by
      result = self.results
      profile = @tool.profile
      if !result.empty?
        sel = Sketchup.active_model.selection
        p_members = Sketchup.active_model.active_entities.find_all{ |e| PBAPI::Classify.profile_member?(e) }
        p_members = p_members.collect{ |e| PBAPI::ProfileMember.new(e) }
        if result[0]
          p_members = p_members.find_all{ |pm| pm.profile.name == profile.name }
        end
        if result[1]
          p_members = p_members.find_all{ |pm| pm.rotation.to_f == profile.rotation.to_f }
        end
        if result[2]
          p_members = p_members.find_all{ |pm| pm.mirror == profile.mirror }
        end
        if result[3]
          p_members = p_members.find_all{ |pm| pm.placement_point == profile.placement_point }
        end
        if result[4]
          p_members = p_members.find_all{ |pm| pm.x_offset.to_f == profile.x_offset.to_f }
        end
        if result[5]
          p_members = p_members.find_all{ |pm| pm.y_offset.to_f == profile.y_offset.to_f }
        end
        if result[6]
          p_members = p_members.find_all{ |pm| pm.material_name == profile.material_name }
        end
        if result[7]
          p_members = p_members.find_all{ |pm| pm.layer_name == profile.layer_name }
        end
        if result[8]
          p_members = p_members.find_all{ |pm| pm.junction_style == profile.junction_style }
        end
        if result[9]
          p_members = p_members.find_all{ |pm| pm.extrude_mode == profile.extrude_mode }
        end
        p_members.each{ |pm| sel.add(pm.instance) }
      end  #end if not result
    end

    def results
      @results
    end

    def refresh_dialog
      values = Sketchup.active_model.get_attribute('ProfileBuilder', 'PMPSValues', nil)
      if values
        values = values.split("|")
        values = values.collect{ |val| val.to_s }
        checkboxes = %w(
          profile
          rotation
          mirror
          placementPoint
          xOffset
          yOffset
          material_name
          layer_name
          junctionStyle
          extrudeMode
        )
        checkboxes.each_index{ |i| setCheckbox(checkboxes[i], values[i]) }
      end
      if @mode == "edit"
        call_js("setEditButtonLabel")
      else
        call_js("setSelectButtonLabel")
      end
    end

  end  #class end
end #module

